# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..artefact import Analysis
from .abc import *
from ..spt_data.abc import SPTData
from ..spt_data import _normalize
from ..roi.abc import ROI
from ..time.abc import Time
import os
import sys
import time
import multiprocessing
import subprocess
import tempfile
import shutil
import glob
import traceback
from tramway.core.hdf5.store import load_rwa
from tramway.core.analyses.base import append_leaf


def join_arguments(args):
    _args = []
    args = iter(args)
    while True:
        try:
            arg = next(args)
        except StopIteration:
            break
        if arg.startswith('--'):
            try:
                key, val = arg[2:].split('=')
            except ValueError:
                return []
            if val[0] == '"':
                cont = []
                while val[-1] != '"':
                    try:
                        val = next(args)
                    except StopIteration:
                        return []
                    cont.append(val)
                if cont:
                    arg = ' '.join([arg]+cont)
            _args.append(arg)
        else:
            return []
    return _args


class Env(AnalyzerNode):
    """
    Implements parts of classes suitable for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.env` main attribute.

    See :class:`LocalHost` or :class:`SlurmOverSSH` for examples of concrete classes.
    """
    __slots__ = ('_interpreter','_script','_working_directory','_worker_count',
            '_pending_jobs','_selectors','_selector_classes','_temporary_files',
            '_collectibles','debug')
    def __init__(self, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._interpreter = 'python3'
        self._script = None
        self._selectors = None
        self._selector_classes = {}
        self._temporary_files = []
        self._working_directory = None
        self.collectibles = None
        self._worker_count = None
        self.pending_jobs = []
        self.debug = False
    @property
    def working_directory(self):
        """
        *str*: Path of the working directory on the worker side
        """
        return self._working_directory
    @working_directory.setter
    def working_directory(self, wd):
        self._working_directory = wd
    @property
    def wd(self):
        return self.working_directory
    wd.__doc__ = working_directory.__doc__
    @wd.setter
    def wd(self, wd):
        self.working_directory = wd
    @property
    def wd_is_available(self):
        """
        *bool*: :const:`True` if the working directory is ready on the worker side
        """
        return True
    @property
    def interpreter(self):
        """
        *str*: Interpreter command on the worker side
        """
        return self._interpreter
    @interpreter.setter
    def interpreter(self, cmd):
        self._interpreter = cmd
    @property
    def script(self):
        """
        *str*: Path to the local script to be executed; meaningful on the submit side only
        """
        return self._script
    @script.setter
    def script(self, file):
        if file is None:
            self._script = None
        elif file.startswith('/'):
            self._script = file
        else:
            self._script = os.path.abspath(os.path.expanduser(file))
    @property
    def worker_count(self):
        """
        *int*: Desired number of workers
        """
        return self._worker_count
    @worker_count.setter
    def worker_count(self, wc):
        self._worker_count = wc
    @property
    def wc(self):
        return self.worker_count
    wc.__doc__ = worker_count.__doc__
    @wc.setter
    def wc(self, wc):
        self.worker_count = wc
    @property
    def pending_jobs(self):
        """
        *list* of *str*: Specification of the jobs to be submitted
        """
        return self._pending_jobs
    @pending_jobs.setter
    def pending_jobs(self, jobs):
        self._pending_jobs = jobs
    @property
    def selectors(self):
        """
        *dict*: Wrapper classes for main :class`~tramway.analyzer.RWAnalyzer` attributes
        """
        return self._selectors
    @selectors.setter
    def selectors(self, sel):
        self._selectors = sel
    @property
    def temporary_files(self):
        """
        *list* of *str*: Temporary files generated on the local side
        """
        return self._temporary_files
    @property
    def analyzer(self):
        return self._parent
    @property
    def collectibles(self):
        return self._collectibles
    @collectibles.setter
    def collectibles(self, cs):
        if cs:
            self._collectibles = set(cs)
        else:
            self._collectibles = set()
    def pending_collectibles(self):
        """
        *set* of *str*: Names or paths of files to be retrieved from the worker side to the submit side
        """
        return self.collectibles
    @property
    def current_stage(self):
        """
        *int* or *list* of *int*: Index of the current stage; meaningful on the submit side only
        """
        assert self.worker_side
        return self.selectors['stage_index']
    def early_setup(self, *argv):
        """
        Determines which side is running and sets the `submit_side`/`worker_side` attributes.

        Takes command-line arguments (``sys.argv``).
        """
        assert argv
        # join arguments with spaces
        args = join_arguments(argv[1:])
        #
        valid_keys = set(('stage-index', 'source', 'region-index', 'segment-index', 'cell-index',
            'working-directory'))
        valid_arguments = {}
        for arg in args:
            if arg.startswith('--'):
                try:
                    key, val = arg[2:].split('=')
                except ValueError:
                    valid_arguments = None
                else:
                    if key in valid_keys:
                        key = key.replace('-', '_')
                        vals = val.split(',')
                        val, vals = vals[0], vals[1:]
                        if val[0]=='"' and val[-1]=='"':
                            val = val[1:-1]
                            _vals = [val]
                            for v in vals:
                                if v[0]=='"' and v[-1]=='"':
                                    _vals.append(v[1:-1])
                                else:
                                    valid_arguments = None
                                    break
                        else:
                            try:
                                val = int(val)
                            except ValueError:
                                valid_arguments = None
                                break
                            _vals = [val]
                            for v in vals:
                                try:
                                    _vals.append(int(v))
                                except ValueError:
                                    valid_arguments = None
                                    break
                        if vals:
                            val = tuple(_vals)
                        valid_arguments[key] = val
                    else:
                        valid_arguments = None
            else:
                valid_arguments = None
            if valid_arguments is None:
                break
        if valid_arguments:
            # worker side
            _valid_arguments = dict(valid_arguments) # copy
            self.wd = _valid_arguments.pop('working_directory', self.wd)
            self.selectors = _valid_arguments
        return valid_arguments
    def setup(self, *argv, create_working_directory=True):
        """
        Determines which side is running and alters iterators of the main
        :class:`~tramway.analyzer.RWAnalyzer` attributes.

        Takes command-line arguments (``sys.argv``).
        """
        valid_arguments = self.early_setup(*argv)
        if valid_arguments:
            # worker side
            self.wd = valid_arguments.pop('working_directory', self.wd)
            self.selectors = valid_arguments
            #self.logger.debug('the following selectors apply to the current job:\n\t{}'.format(self.selectors))
            #
            if self.script is not None and self.script.endswith('.ipynb'):
                self.script = self.script[:-5]+'py'
            #
            try:
                sources = valid_arguments['source']
            except KeyError:
                pass
            else:
                if isinstance(sources, str):
                    sources = os.path.expanduser(sources)
                else:
                    sources = tuple([
                        os.path.expanduser(source) for source in sources
                        ])
                self.selectors['source'] = sources
                #self.logger.debug('selecting source: '+', '.join((sources,) if isinstance(sources, str) else sources))
            #
            for f in self.analyzer.spt_data: # TODO: check why self.spt_data_selector(..) does not work
                f._analyses.rwa_file = self.make_temporary_file(suffix='.rwa', output=True)
                f._analyses.autosave = True
        elif self.script is None:
            # not tested!
            candidate_files = [ f for f in os.listdir() \
                    if f.endswith('.py') or f.endswith('.ipynb') ]
            if candidate_files and not candidate_files[1:]:
                self.script = candidate_files[0]
                self.logger.info('candidate script: {} (in {})'.format(self.script, os.getcwd()))
            raise ValueError('attribute `script` is not set')
        elif create_working_directory:
            self.make_working_directory()
            self.logger.info('working directory: '+self.wd)
    def spt_data_selector(self, spt_data_attr):
        """
        Wraps the :attr:`~tramway.analyzer.RWAnalyzer.spt_data` attribute.
        """
        if isinstance(spt_data_attr, Initializer) or isinstance(spt_data_attr, Proxy):
            return spt_data_attr
        cls = type(spt_data_attr)
        try:
            selector_cls = self._selector_classes[cls]
        except KeyError:
            try:
                sources = self.selectors['source']
            except KeyError:
                return spt_data_attr
            if isinstance(sources, str):
                sources = (sources,)
            #logger = self.logger
            try:
                alias = all([ bool(f.alias) for f in self._eldest_parent.spt_data ])
            except AttributeError:
                alias = False
            if alias:
                aliases = sources # TODO: this can be checked
                class selector_cls(Proxy):
                    __slots__ = ()
                    def __iter__(self):
                        for f in self._proxied:
                            if f.alias in aliases:
                                #logger.debug('source {} selected'.format(f.source))
                                yield f
            else:
                sources = set([ _normalize(_s) for _s in sources ])
                class selector_cls(Proxy):
                    __slots__ = ()
                    def __iter__(self):
                        for f in self._proxied:
                            if _normalize(f.source) in sources:
                                #logger.debug('source {} selected'.format(f.source))
                                yield f
            SPTData.register(selector_cls)
            self._selector_classes[cls] = selector_cls
        return selector_cls(spt_data_attr)
    def roi_selector(self, roi_attr):
        """
        Wraps the :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute and/or the individual
        :attr:`..spt_data.SPTDataItem.roi` attributes.
        """
        if isinstance(roi_attr, Initializer) or isinstance(roi_attr, Proxy):
            return roi_attr
        cls = type(roi_attr)
        try:
            selector_cls = self._selector_classes[cls]
        except KeyError:
            try:
                region = self.selectors['region_index']
            except KeyError:
                return roi_attr
            regions = set([region]) if isinstance(region, int) else set(region)
            def _region(index_arg):
                if index_arg is None:
                    return region
                elif callable(index_arg):
                    return lambda r: r in regions and index_arg(r)
                elif index_arg in regions:
                    return index_arg # or lambda r: True
                else:
                    return lambda r: False
            class selector_cls(Proxy):
                __slots__ = ()
                def as_support_regions(self, index=None, source=None, return_index=False):
                    yield from self._proxied.as_support_regions(_region(index), source, return_index)
                def as_individual_roi(self, *args, **kwargs):
                    raise NotImplementedError
            ROI.register(selector_cls)
            self._selector_classes[cls] = selector_cls
        return selector_cls(roi_attr)
    def time_selector(self, time_attr):
        """
        Wraps the :attr:`~tramway.analyzer.RWAnalyzer.time` attribute.
        """
        if isinstance(time_attr, Initializer) or isinstance(time_attr, Proxy):
            return time_attr
        cls = type(time_attr)
        try:
            segment = self.selectors['segment_index']
        except KeyError:
            return time_attr
        else:
            segments = set([segment]) if isinstance(segment, int) else set(segment)
            def _segment(index_arg):
                if index_arg is None:
                    return segment
                elif callable(index_arg):
                    return lambda t: t in segments and index_arg(t)
                elif index_arg in segments:
                    return index_arg # or lambda t: True
                else:
                    return lambda t: False
            logger = self.logger
            class selector_cls(Proxy):
                __slots__ = ()
                def as_time_segments(self, sampling, maps=None, index=None, return_index=False, return_times=True):
                    yield from self._proxied.as_time_segments(sampling, maps, _segment(index), return_index, return_times)
            Time.register(selector_cls)
            self._selector_classes[cls] = selector_cls
        return selector_cls(time_attr)
    @property
    def submit_side(self):
        """
        *bool*: :const:`True` if currently running on the submit side
        """
        return self.selectors is None
    @property
    def worker_side(self):
        """
        *bool*: :const:`True` if currently running on the worker side
        """
        return self.selectors is not None
    def make_working_directory(self):
        """
        """
        assert self.submit_side
        self.wd = self.make_temporary_file(directory=True)
    def make_temporary_file(self, output=False, directory=False, **kwargs):
        """
        Arguments:

            output (bool): do not delete the file once the task is done.

            directory (bool): make a temporary directory.

        More keyword arguments can be passed to :func:`tempfile.mkstemp`.
        See for example *suffix*.
        """
        if self.wd_is_available:
            parent_dir = self.working_directory
        else:
            parent_dir = None # standard /tmp location
        if directory:
            tmpfile = tempfile.mkdtemp(dir=parent_dir)
        else:
            fd, tmpfile = tempfile.mkstemp(dir=parent_dir, **kwargs)
            os.close(fd)
        if not output:
            self._temporary_files.append(tmpfile)
        return tmpfile
    def dispatch(self, **kwargs):
        """
        Prepares the worker side. To be called from the submit side only.
        """
        if not kwargs:
            self.prepare_script()
            return True
    def __del__(self):
        #self.delete_temporary_data()
        pass
    def close(self):
        self._parent.spt_data.close()
    def delete_temporary_data(self):
        """
        Deletes all the temporary data, on both the submit and worker sides.
        """
        for file in self._temporary_files[::-1]:
            if os.path.isdir(file):
                try:
                    shutil.rmtree(file)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception as e:
                    self.logger.warning('temporary files removal failed with the following error:\n{}'.format(e))
            elif os.path.isfile(file):
                try:
                    os.unlink(file)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception as e:
                    self.logger.debug('temporary file removal failed with the following error:\n{}'.format(e))
        self._temporary_files = []
    def make_job(self, stage_index=None, source=None, region_index=None, segment_index=None):
        """
        Registers a new pending job.

        Arguments:

            stage_index (*int* or *list* of *int*): stage index(ices)

            source (str): SPT datablock identifier (source path or alias)

            region_index (int): index of the support region
                (see also :meth:`~tramway.analyzer.roi.ROI.as_support_regions`)

            segment_index (int): index of the time segment

        """
        assert self.submit_side
        command_options = ['--working-directory="{}"'.format(self.wd)]
        if isinstance(stage_index, list):
            command_options.append(','.join(['--stage-index={}']+['{}']*(len(stage_index)-1)).format(*stage_index))
        elif stage_index is not None:
            command_options.append('--stage-index={:d}'.format(stage_index))
        if source is not None:
            command_options.append('--source="{}"'.format(source))
        if region_index is not None:
            command_options.append('--region-index={:d}'.format(region_index))
        if segment_index is not None:
            command_options.append('--segment-index={:d}'.format(segment_index))
        self.pending_jobs.append(tuple(command_options))
    @classmethod
    def _combine_analyses(cls, wd, data_location, logger, *args, directory_mapping={},
            inplace=False, reload_existing_rwa_files=False):
        """
        Loads the generated interim rwa files, combines them and returns the list
        of the resulting files.

        Should be run on the worker side only, where the collectible files are
        available.
        """
        # read the interim files and combine the analysis trees
        analyses, original_files = {}, {}
        output_files = glob.glob(os.path.join(wd, '*.rwa'))
        loaded_files = []
        while output_files:
            output_file = output_files.pop()
            if os.stat(output_file).st_size == 0:
                logger.info('skipping empty file '+output_file)
                os.unlink(output_file)
                continue
            logger.info('reading file: {}'.format(output_file))
            try:
                __analyses = load_rwa(output_file,
                        lazy=True, force_load_spt_data=False)
            except:
                logger.critical(traceback.format_exc())
                #raise
                continue
            try:
                source = __analyses.metadata['datafile']
                if source is None:
                    raise KeyError
            except KeyError:
                logger.debug(str(__analyses))
                logger.debug('metadata: ',__analyses.metadata)
                logger.critical('key `datafile` undefined in metadata')
                #analyses = {}; break
                continue
            try:
                _analyses = analyses[source]
            except KeyError:
                analyses[source] = __analyses
                original_files[source] = [output_file]
            else:
                try:
                    Analysis.rebase_tree(_analyses, __analyses)
                except ValueError:
                    print(_analyses)
                    print(__analyses)
                    raise
                else:
                    original_files[source].append(output_file)
            loaded_files.append(output_file)

        # write all the combined trees in output files
        end_result_files = []
        for source in analyses:
            logger.info('for source file: {}...'.format(source))

            # TODO: if isinstance(spt_data, (StandaloneRWAFile, RWAFiles))
            #       pass the input rwa file paths to _combine_analyses;
            #       this would be helpful for user-defined rwa filenames
            #       that do not match the original SPT data filename.

            # determine the output rwa filename
            rwa_file = os.path.splitext(os.path.normpath(source))[0]+'.rwa'
            if inplace:
                if original_files[source][1:]:
                    if inplace:
                        fd, rwa_file = tempfile.mkstemp(dir=wd, suffix='.rwa')
                        os.close(fd)
                else:
                    rwa_file = None
            elif os.path.isabs(rwa_file):
                if directory_mapping:
                    for to_be_replaced in directory_mapping:
                        if rwa_file.startswith(to_be_replaced):
                            rwa_file = rwa_file[len(to_be_replaced):]
                            replacement = directory_mapping[to_be_replaced]
                            if replacement:
                                rwa_file = replacement+rwa_file # and NOT os.path.join, since rwa_file may start with '/'
            elif not os.path.isabs(os.path.expanduser(rwa_file)) and data_location:
                rwa_file = os.path.join(data_location, rwa_file)

            # log the outcome
            if rwa_file is None:
                # designate the standalone .rwa file as an output file
                # so that it is spared while cleaning up the working directory
                rwa_file = original_files[source][0]
                #
                logger.info('interim file left untouched: {}'.format(rwa_file))
            else:
                logger.info('writing file: {}'.format(rwa_file))

            # flush the combined data
            if original_files[source][1:]:
                assert analyses[source] is not None
                try:
                    Analysis.save(os.path.expanduser(rwa_file), analyses[source], force=True,
                            rebase=reload_existing_rwa_files)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except:# FileNotFoundError:
                    if inplace:
                        raise
                    logger.warning('writing file failed: {}'.format(rwa_file))
                    local_rwa_file = rwa_file
                    fd, remote_rwa_file = tempfile.mkstemp(dir=wd, suffix='.rwa')
                    os.close(fd)
                    logger.info('writing file: {}...'.format(remote_rwa_file))
                    Analysis.save(remote_rwa_file, analyses[source], force=True,
                            rebase=reload_existing_rwa_files)
                    rwa_file = (remote_rwa_file, local_rwa_file)
            elif not inplace:
                if analyses[source]._data is not None:
                    analyses[source]._data.store.close() # close the file descriptor
                original_file = original_files[source][0]
                # copy original_file to rwa_file
                try:
                    with open(rwa_file, 'wb') as o:
                        with open(original_file, 'rb') as i:
                            o.write(i.read())
                except OSError:
                    remote_rwa_file = original_file
                    local_rwa_file = rwa_file
                    rwa_file = (remote_rwa_file, local_rwa_file)

            # register the output files
            end_result_files.append(rwa_file)

        # delete the input interim files
        to_keep = [ f[0] if isinstance(f,tuple) else f for f in end_result_files ]
        for output_file in loaded_files:
            if output_file not in to_keep:
                try:
                    os.unlink(output_file)
                except FileNotFoundError:
                    pass
                except PermissionError as e:
                    if os.name == 'nt':
                        logger.debug(str(e))
                    else:
                        raise
        #
        return end_result_files
    def collect_results(self, stage_index=None, inplace=False, reload_existing_rwa_files=False):
        """
        Calls :meth:`_combine_analyses` for the remote data location and the
        current working directory and stage index,
        and retrieves the combined files from the worker side to the submit side,
        if they are different hosts.

        Returns :const:`True` if files are collected/retrieved.

        If `inplace` is :const:`True`, user-defined collectibles are disregarded
        and only the *.rwa* files are sought for and combined.
        In addition, no files are moved outside the working directory on the
        worker side.
        The intermediate *.rwa* files are combined into new intermediate *.rwa* files.
        This is useful to maintain fewer intermediate files and free significant amount
        of disk space.
        """
        return bool(self._combine_analyses(self.wd, None, self.logger, stage_index, inplace=inplace,
                reload_existing_rwa_files=reload_existing_rwa_files))
    def prepare_script(self, script=None):
        main_script = script is None
        if main_script:
            script = self.script
        # load
        if script.endswith('.ipynb'):
            content = self.import_ipynb(script)
        else:
            with open(self.script, 'r') as f:
                content = f.readlines()
        # modify
        filtered_content = self.filter_script_content(content)
        if script in self.temporary_files:
            tmpfile = script # reuse file
        else:
            tmpfile = self.make_temporary_file(suffix='.py', text=True)
        # flush
        with open(tmpfile, 'w') as f:
            for line in filtered_content:
                f.write(line)
        if main_script:
            self.script = tmpfile
        else:
            return tmpfile
    def filter_script_content(self, content):
        r"""
        Processes the script content for its dispatch onto the worker side.

        Arguments:

            content (*list* of *str*): lines with the :const:`'\n'` character at the end of each line

        Returns:

            *list* of *str*: modified lines

        """
        filtered_content = []
        for line in content:
            if line.startswith('#'):
                pass
            elif line.startswith('get_ipython('):
                continue
            elif '.run()' in line:
                # last line
                filtered_content.append(line)
                break
            filtered_content.append(line)
        return filtered_content
    def import_ipynb(self, notebook):
        """
        Extracts the code content of a IPython notebook.

        Arguments:

            notebook (str): path of the *.ipynb* notebook file

        Returns:

            *list* of *str*: extracted lines

        """
        cmd = 'jupyter nbconvert --to python "{}" --stdout'.format(notebook)
        self.logger.info('running: '+cmd)
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            if not isinstance(out, str):
                out = out.decode('utf-8')
            content = out.splitlines(True)
        else:
            content = None
        if err:
            if not isinstance(err, str):
                err = err.decode('utf-8')
            if err != '[NbConvertApp] Converting notebook {} to python\n'.format(notebook):
                self.logger.error(err)
        return content
    def interrupt_jobs(self):
        """
        Interrupts the running jobs.
        """
        pass


class LocalHost(Env):
    """
    Runs jobs in local **python** processes.
    """
    __slots__ = ('running_jobs',)
    def __init__(self, **kwargs):
        Env.__init__(self, **kwargs)
        self.interpreter = sys.executable
        self.running_jobs = []
        self.wc = None
    @property
    def worker_count(self):
        return self._worker_count
    @worker_count.setter
    def worker_count(self, wc):
        if wc is None:
            wc = max(1, multiprocessing.cpu_count() - 1)
        elif wc < 0:
            wc = max(1, multiprocessing.cpu_count() + wc)
        self._worker_count = wc
    def submit_jobs(self):
        assert self.submit_side
        self.running_jobs = []
        for j,job in enumerate(self.pending_jobs):
            if len(self.running_jobs) == self.wc:
                self.wait_for_job_completion(1)
            self.logger.debug('submitting: '+( ' '.join(['{}']*(len(job)+2)).format(self.interpreter, self.script, *job) ))
            p = subprocess.Popen([self.interpreter, self.script, *job],
                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            self.running_jobs.append((j,p))
        self.pending_jobs = []
    def wait_for_job_completion(self, count=None):
        assert self.submit_side
        n = 0
        for j,p in self.running_jobs:
            out, err = p.communicate()
            out, err = out.rstrip(), None if err is None else err.rstrip()
            if out:
                if not isinstance(out, str):
                    out = out.decode('utf-8')
                self.logger.info(out)
            if err:
                if not isinstance(err, str):
                    err = err.decode('utf-8')
                self.logger.error(err)
            self.logger.debug('job {:d} done\n'.format(j))
            n += 1
            if n==count:
                self.running_jobs = self.running_jobs[n:]
                return
        self.running_jobs = []
    def interrupt_jobs(self):
        for j,p in self.running_jobs:
            p.terminate()
        for j,p in self.running_jobs:
            out, err = p.communicate()
            if err:
                if not isinstance(err, str):
                    err = err.decode('utf-8')
                self.logger.error(err)
                return False
            elif out:
                if not isinstance(out, str):
                    out = out.decode('utf-8')
                self.logger.info(out)
        self.running_jobs = []
        return True

Environment.register(LocalHost)


class UpdatedDict(dict):
    __slots__ = ('_proper', '_inherited')
    def __init__(self, proper, shared, shared_keys=None):
        self._proper = proper
        if shared_keys:
            self._inherited = { k: shared[k] for k in shared if k in shared_keys }
        else:
            self._inherited = dict(shared)
        self._inherited.update(self._proper)
    def __repr__(self):
        return repr(self._inherited)
    def __str__(self):
        return str(self._inherited)
    def __contains__(self, key):
        return key in self._inherited
    def __iter__(self):
        return iter(self._inherited)
    def __len__(self):
        return len(self._inherited)
    def __getitem__(self, key):
        return self._inherited[key]
    def __setitem__(self, key, value):
        self._proper[key] = value
        self._inherited[key] = value
    def __delitem__(self, key):
        del self._inherited[key]
        try:
            del self._proper[key]
        except KeyError:
            pass
    def clear(self):
        self._proper.clear()
        self._inherited = {}
    def copy(self):
        return self._inherited.copy()
    def get(self, *args):
        return self._inherited.get(*args)
    def items(self):
        return self._inherited.items()
    def keys(self):
        return self._inherited.keys()
    def pop(self, key, *args):
        try:
            value = self._inherited.pop(key)
        except KeyError:
            if args:
                value = args[0]
            else:
                raise
        else:
            try:
                del self._proper[key]
            except KeyError:
                pass
        return value
    def popitem(self):
        key, value = self._inherited.popitem()
        try:
            del self._proper[key]
        except KeyError:
            pass
        return key, value
    def setdefault(self, key, default=None):
        try:
            value = self._inherited[key]
        except KeyError:
            value = self._proper[key] = self._inherited[key] = default
        return value
    def update(self, other):
        self._proper.update(other)
        self._inherited.update(other)
    def values(self):
        return self._inherited.values()


class Slurm(Env):
    """
    Not supposed to properly run, as TRamWAy is expected to be called
    inside a container;
    see :class:`SlurmOverSSH` instead.
    """
    __slots__ = ('_sbatch_options','_job_id','refresh_interval','_srun_options')
    def __init__(self, **kwargs):
        Env.__init__(self, **kwargs)
        self._sbatch_options = dict(
                output='%J.out',
                error='%J.err',
                )
        self._srun_options = {}
        self._job_id = None
        self.refresh_interval = 20
    @property
    def sbatch_options(self):
        return self._sbatch_options
    @property
    def srun_options(self):
        return UpdatedDict(self._srun_options, self.sbatch_options, self.slurm_common_options)
    @srun_options.setter
    def srun_options(self, options):
        if options is None:
            self._srun_options = {}
        elif isinstance(options, UpdatedDict):
            self._srun_options = options._proper
        elif isinstance(options, dict):
            self._srun_options = options
        else:
            raise TypeError("wrong type for srun_options: '{}'".format(type(options).__name__))
    @property
    def slurm_common_options(self):
        """
        tuple:
            set of options that are common to sbatch and srun;
            if a "common" option found in :attr:`sbatch_options`
            and not in :attr:`srun_options`, :attr:`srun_options`
            is updated with this option
        """
        return ('p', 'partition', 'q', 'qos')
    @property
    def job_name(self):
        try:
            name = self.sbatch_options['job-name']
        except KeyError:
            try:
                name = self.sbatch_options['J']
            except KeyError:
                name = self.sbatch_options['job-name'] = os.path.splitext(os.path.basename(self.script))[0]
            else:
                self.sbatch_options['job-name'] = name
        return name
    @job_name.setter
    def job_name(self, name):
        self.sbatch_options['job-name'] = name
    @property
    def job_id(self):
        return self._job_id
    @job_id.setter
    def job_id(self, i):
        self._job_id = i
    @property
    def task_count(self):
        return self.worker_count
    @task_count.setter
    def task_count(self, count):
        self.worker_count = count
    def setup(self, *argv, **kwargs):
        Env.setup(self, *argv, **kwargs)
        if self.submit_side:
            self.job_name # sets job-name sbatch option
            output_log = self.sbatch_options['output']
            if not os.path.isabs(os.path.expanduser(output_log)):
                output_log = '/'.join((self.wd, output_log)) # NOT os.path.join
                self.sbatch_options['output'] = output_log
            error_log = self.sbatch_options['error']
            if not os.path.isabs(os.path.expanduser(error_log)):
                error_log = '/'.join((self.wd, error_log))
                self.sbatch_options['error'] = error_log
    def make_sbatch_script(self, stage=None, path=None):
        assert self.submit_side
        if path is None:
            sbatch_script = self.make_temporary_file(suffix='.sh' if stage is None else '-stage{:d}.sh'.format(stage), text=True)
        else:
            sbatch_script = path
        self.job_name # set default job name if not defined yet
        with open(sbatch_script, 'w', encoding='utf-8', newline='\n') as f:
            f.write('#!/bin/bash\n')
            for option, value in self.sbatch_options.items():
                if option[1:]:
                    line = '#SBATCH --{}={}\n'
                else:
                    line = '#SBATCH -{} {}\n'
                if isinstance(value, str) and ' ' in value:
                    value = '"{}"'.format(value)
                f.write(line.format(option, value))
            f.write('#SBATCH --array=0-{:d}{}\n'.format(
                len(self.pending_jobs)-1,
                '' if self.wc is None else '%{:d}'.format(self.wc)))
            f.write('\ndeclare -a tasks\n')
            wd, si = None, None
            for j, job in enumerate(self.pending_jobs):
                assert job[0].startswith('--working-directory=')
                if wd is None:
                    assert j == 0
                    wd = job[0]
                else:
                    assert wd == job[0]
                if job[1].startswith('--stage-index='):
                    if si is None:
                        assert j == 0
                        si = job[1]
                    else:
                        assert si == job[1]
                    job = job[2:]
                else:
                    assert si is None
                    job = job[1:]
                f.write('tasks[{:d}]="{}"\n'.format(j,
                    ' '.join(['{}']*len(job)).format(*job).replace('"',r'\"')))
            if si is None:
                common_task_args = wd
            else:
                common_task_args = '{} {}'.format(wd.replace('"',r'\"'), si)
            #f.write('\npushd "{}" > /dev/null\n'.format(self.wd))
            f.write('\n{} {} {} ${{tasks[SLURM_ARRAY_TASK_ID]}}\n'.format(self.interpreter, self.script,
                common_task_args))
            #f.write('\npopd > /dev/null\n')
            f.write('\nunset tasks\n')
        if True:
            with open(sbatch_script, 'r') as f:
                self.logger.debug(f.read())
        return sbatch_script
    def submit_jobs(self):
        # this code below does not actually run as the :meth:`submit_jobs` method is overwritten
        # in :class:`SlurmOverSSH`
        sbatch = 'sbatch'
        sbatch_script = self.make_sbatch_script()
        self.logger.info('running: {} {}'.format(sbatch, sbatch_script))
        p = subprocess.Popen([sbatch, sbatch_script],
                stderr=subprocess.STDOUT, encoding='utf-8')#, stdout=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            self.logger.info(out)
            if out.startswith('Submitted batch job '):
                self.job_id = out.strip().split(' ')[-1]
        if err:
            self.logger.error(err)
        self.pending_jobs = []
    def wait_for_job_completion(self):
        prefix = '{}_['.format(self.job_id)
        try:
            while True:
                p = subprocess.Popen(('squeue', '-j '+self.job_id, '-h', '-o "%.18i %.2t %.10M %R"'),
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
                out, err = p.communicate()
                if err:
                    self.logger.error(err)
                elif out:
                    for row in out.splitlines():
                        if row.startswith(prefix):
                            break
                    out = row.split(' ')
                    status, time_used, reason = out[1], out[2], ' '.join(out[3:])
                    self.logger.info('status: {}   time used: {}   reason: {}'.format(status, time_used, reason))
                else:
                    break
                time.sleep(self.refresh_interval)
        except:
            self.logger.info('killing jobs with: scancel '+self.job_id)
            subprocess.Popen(('scancel', self.job_id)).communicate()
            raise


__remote_host_attrs__ = ('_ssh','_local_data_location','_remote_data_location','_directory_mapping')

class RemoteHost(object):
    """
    mixin class.

    children classes should define attributes listed in `__remote_host_attrs__`.

    minimum base implementation:

    .. code-block:: python

        assert issubclass(cls, Env)

        class Cls(cls, RemoteHost):
            __slots__ = __remote_host_attrs__
            def __init__(self, **kwargs):
                cls.__init__(self, **kwargs)
                RemoteHost.__init__(self)
            wd_is_available = RemoteHost.wd_is_available
            make_working_directory = RemoteHost.make_working_directory
            def setup(self, *argv, **kwargs):
                cls.setup(self, *argv, **kwargs)
                RemoteHost.setup(self, *argv)
            def filter_script_content(self, content):
                return RemoteHost.filter_script_content(self,
                        cls.filter_script_content(self, content))
            def make_job(self, stage_index=None, source=None, region_index=None,
                        segment_index=None):
                cls.make_job(self, stage_index, self.format_source(source),
                        region_index, segment_index)
            def collect_results(self, stage_index=None, inplace=False,
                        reload_existing_rwa_files=False):
                log_pattern = '*.out' # to be adapted
                RemoteHost.collect_results(self, log_pattern, stage_index,
                    inplace=inplace, reload_existing_rwa_files=reload_existing_rwa_files)
            def delete_temporary_data(self):
                RemoteHost.delete_temporary_data(self)
                cls.delete_temporary_data(self)

    """
    __slots__ = ()
    def __init__(self):
        self._ssh = None
        self._local_data_location = None
        self._remote_data_location = None
        self._directory_mapping = {}
    @property
    def ssh(self):
        """
        SSH communication interface.

        See also :class:`~tramway.analyzer.env.ssh.Client`.
        """
        if self._ssh is None:
            from tramway.analyzer.env import ssh
            self._ssh = ssh.Client()
        return self._ssh
    @property
    def remote_data_location(self):
        """
        Data location on the remote host (worker side).

        If defined, the current directory will be changed to this location
        on the remote host.
        """
        return self._remote_data_location
    @remote_data_location.setter
    def remote_data_location(self, path):
        former_location = self.remote_data_location
        if path != former_location:
            if self.local_data_location is not None:
                if path is None: # former_location is not None
                    # therefore key `local_data_location` can be found in `directory_mapping`
                    del self.directory_mapping[self.local_data_location]
                else:
                    self.directory_mapping[self.local_data_location] = path
            self._remote_data_location = path
    @property
    def local_data_location(self):
        """
        Data location on the local host (submit side).
        """
        return self._local_data_location
    @local_data_location.setter
    def local_data_location(self, path):
        former_location = self.local_data_location
        if path != former_location:
            if self.remote_data_location is not None:
                if path is None: # former_location is not None
                    del self.directory_mapping[former_location]
                else:
                    self.directory_mapping[path] = self.remote_data_location
            self._local_data_location = path
    @property
    def directory_mapping(self):
        """
        Directory mapping with local paths as keys
        and the corresponding remote paths as values.

        Paths should not have slashes at the end.
        """
        return self._directory_mapping
    @property
    def collection_interpreter(self):
        return self.interpreter
    @property
    def wd_is_available(self):
        return self.worker_side
    def make_working_directory(self):
        if self.submit_side:
            cmd = 'mkdir -p "{}"; mktemp -d -p "{}"'.format(self.working_directory, self.working_directory)
            out, err = self.ssh.exec(cmd)
            if err:
                # attribute `logger` must be defined by the concrete parent class
                self.logger.error(err)
            elif out:
                self.working_directory = out.rstrip()
        else:
            assert os.path.isdir(self.wd)
    def setup(self, *argv):
        if self.worker_side:
            self.script = '/'.join((self.wd, os.path.basename(self.script)))
    @classmethod
    def _format_collectibles(cls, paths):
        home = os.path.expanduser('~')
        collectibles = []
        for path in paths:
            mapping = isinstance(path, tuple)
            if mapping:
                src_path, path = path
            path = os.path.normpath(path)
            if os.path.isabs(path) and path.startswith(home):
                path = '~'+path[len(home):]
            if mapping:
                path = '{{{0}=>{1}}}'.format(src_path, path)
            collectibles.append(path)
        return collectibles
    def format_collectibles(self, paths=None):
        """
        Calls `_format_collectibles` with the list of registered collectibles.
        """
        if paths is None:
            paths = self.collectibles
        return self._format_collectibles(paths)
    def format_source(self, source):
        if source is not None:
            source = os.path.normpath(source)
            home = os.path.expanduser('~')
            if os.path.isabs(source) and source.startswith(home):
                source = '~'+source[len(home):]
            for _dir in self.directory_mapping:
                if source.startswith(_dir):
                    source = self.directory_mapping[_dir]+source[len(_dir):]
                    break
            source = source.replace(os.sep, '/')
        return source
    def dispatch(self, **kwargs):
        if not kwargs:
            self.prepare_script()
            src = self.script
            dest = os.path.basename(self.script)
            #if dest.endswith('.ipynb'):
            #    dest = dest[:-5]+'py'
            dest = '/'.join((self.wd, dest))
            attrs = self.ssh.put(src, dest)
            self.logger.debug(attrs)
            self.script = dest
            self.logger.info('Python script location: '+dest)
            return True
    @classmethod
    def remote_dependencies(cls):
        return ''
    def exec_inline_python(self, inline_code):
        python = []
        if self.remote_dependencies():
            python.append(self.remote_dependencies())
        python.append(self.collection_interpreter)
        python = '; '.join(python)
        return self.ssh.exec('{} -c "{}"'.format(python, inline_code.replace('"', r'\"')), shell=True, logger=self.logger)
    @classmethod
    def _collectible_prefix(cls, stage_index=None):
        """
        Computes the starting substring of the log message that reports
        the list of collectibles.
        """
        if stage_index is None:
            prefix = 'OUTPUT_FILES='
        elif isinstance(stage_index, (tuple, list)):
            prefix = 'OUTPUT_FILES[{}]='.format('-'.join([ str(i) for i in stage_index ]))
        else:
            prefix = 'OUTPUT_FILES[{}]='.format(stage_index)
        return prefix
    def collectible_prefix(self, stage_index=None):
        """
        Calls :meth:`_collectible_prefix` for the current stage(s).
        """
        if stage_index is None:
            # attribute `current_stage` is expected to defined in the concrete parent class
            stage_index = self.current_stage
        return self._collectible_prefix(stage_index)
    def collect_results(self, _log_pattern, stage_index=None, _parent_cls='Env', inplace=False,
            reload_existing_rwa_files=False):
        """
        Downloads the reported collectibles from the remote host (worker side)
        to the local host (submit side).

        See :meth:`Env.collect_results` for information on the distinct behavior
        of the `inplace=True` case.
        """
        _prefix = self.collectible_prefix(stage_index)
        data_loc = self.remote_data_location if self.remote_data_location else ''
        home = os.path.expanduser('~')
        if home not in self.directory_mapping:
            self.directory_mapping[home] = '~'
        code = """\
from tramway.analyzer import environments, BasicLogger

wd = '{}'
data_location = '{}'
logger = BasicLogger()
stage = {!r}
log_pattern = '{}'
directory_mapping = {}

files  = environments.{}._combine_analyses(wd, data_location, logger, stage,
            directory_mapping=directory_mapping, inplace={},
            reload_existing_rwa_files={})\
""".format(self.wd, data_loc, stage_index, _log_pattern, self.directory_mapping, _parent_cls,
            inplace, reload_existing_rwa_files)
        if not inplace:
            code += """
files += environments.RemoteHost._collectibles_from_log_files(wd, log_pattern, stage)

files  = environments.RemoteHost._format_collectibles(files)

print('{}'+';'.join(files))\
""".format(_prefix)
        local_script = self.make_temporary_file(suffix='.py', text=True)
        with open(local_script, 'w') as f:
            f.write(code)
        remote_script = '/'.join((self.wd, os.path.basename(local_script)))
        attrs = self.ssh.put(local_script, remote_script, confirm=True)
        self.logger.debug(attrs)
        log_file = os.path.splitext(remote_script)[0]+'.log'
        remote_deps = self.remote_dependencies()
        if remote_deps and not remote_deps.rstrip().endswith(';'):
            remote_deps = remote_deps+'; '
        cmd = '{}{} {}; rm {}'.format(remote_deps,
                self.collection_interpreter, remote_script, remote_script)
        out, err = self.ssh.exec(cmd, shell=True, logger=self.logger)
        out, err = out.rstrip(), err.rstrip()
        if err:
            if out:
                self.logger.debug(out)
            self.logger.error(err)
        if inplace:
            return False
        if not out:
            return False
        if not err:
            self.logger.debug(out)
        out = out.splitlines()
        while True:
            try:
                line = out.pop() # starting from last line
            except IndexError: # empty list
                raise RuntimeError('missing output: {}...'.format(_prefix)) from None
            if line.startswith(_prefix):
                end_result_files = line[len(_prefix):].split(';')
                break
        any_transfer = False
        for end_result_file in end_result_files:
            if not end_result_file:
                continue
            if end_result_file[0]=='{' and '=>' in end_result_file and end_result_file[-1]=='}':
                src, dest = end_result_file[1:-1].split('=>',1)
            else:
                src = dest = end_result_file
            dest = os.path.normpath(dest)
            self.logger.info('retrieving file: '+dest)
            for local, remote in self.directory_mapping.items():
                if dest.startswith(remote):
                    dest = local+dest[len(remote):]
                    break
            try:
                self.ssh.get(src, dest)
            except FileNotFoundError: # the target file might be empty
                self.logger.warning('failed')
            else:
                any_transfer = True
        return any_transfer
    @classmethod
    def _collectibles_from_log_files(cls, wd, log_pattern, stage_index=None):
        """
        Reads the collectible names reported in the log files.
        """
        collectibles = []
        _prefix = cls._collectible_prefix(stage_index)
        home = os.path.expanduser('~')
        log_files = glob.glob(os.path.join(wd,log_pattern))
        for log_file in log_files:
            with open(log_file, 'r') as f:
                try:
                    last_line = f.readlines()[-1]
                except IndexError:
                    continue
            if last_line.startswith(_prefix):
                last_line = last_line.rstrip()
                for collectible in last_line[len(_prefix):].split(';'):
                    if collectible:
                        collectible = os.path.normpath(collectible)
                        #if os.path.isabs(collectible):
                        #    if collectible.startswith(home):
                        #        collectible = '~'+collectible[len(home):]
                        collectibles.append(collectible)
        return collectibles
    def filter_script_content(self, content):
        filtered_content = []
        if self.remote_data_location:
            filtered_content.append('import os\n')
            filtered_content.append("os.chdir(os.path.expanduser('{}'))\n".format(self.remote_data_location))
            filtered_content.append('\n')
        for line in content:
            pattern = '.spt_data.from_'
            if self.remote_data_location and pattern in line:
                analyzer, suffix = line.split(pattern)
                parts = suffix.split('(')[::-1]
                initializer = parts.pop()
                arg = parts.pop()
                quote = arg[0]
                if quote in "'\"":
                    arg, tail = arg[1:].split(quote, 1)
                    if tail and tail[0] == ')':
                        self.local_data_location, arg = os.path.split(arg)
                        line = '{}{}{}({}{}/{}{})\n'.format(analyzer, pattern, initializer,
                                quote,
                                self.remote_data_location, arg,
                                quote)
            filtered_content.append(line)
        # last line is "[analyzer].run()\n"
        call = line.find('.run()')
        if 0<call:
            analyzer_var = line[:call]
            # insert mapping for home directories
            filtered_content = filtered_content[:-1]
            filtered_content.append("""
if {1}.env.directory_mapping is None:
    {1}.env.directory_mapping = {{}}
if '{0}' not in {1}.env.directory_mapping:
    {1}.env.directory_mapping['{0}'] = '~'
""".format(os.path.expanduser('~').replace('\\',r'\\'), analyzer_var))
            filtered_content.append(line)
            # append output listing
            filtered_content.append(\
                    "\nprint({0}.env.collectible_prefix()+';'.join({0}.env.format_collectibles()))\n".format(
                        analyzer_var))
        # 
        return filtered_content
    def delete_temporary_data(self):
        """
        Deletes the worker-side working directory.
        """
        out, err = self.ssh.exec('rm -rf '+self.wd)
        if out:
            self.logger.info(out)
        if err:
            self.logger.error(err)


class SlurmOverSSH(Slurm, RemoteHost):
    """
    Calls *sbatch* through an SSH connection to a Slurm server.

    .. note::

        All filepaths should be absolute or relative to the $HOME directory.
        This applies especially to all file-based initializers.

    """
    __slots__ = __remote_host_attrs__
    def __init__(self, **kwargs):
        Slurm.__init__(self, **kwargs)
        RemoteHost.__init__(self)
    wd_is_available = RemoteHost.wd_is_available
    make_working_directory = RemoteHost.make_working_directory
    def setup(self, *argv, **kwargs):
        Slurm.setup(self, *argv, **kwargs)
        RemoteHost.setup(self, *argv)
    def dispatch(self, **kwargs):
        if 'stage_options' in kwargs:
            stage_options = kwargs['stage_options']
            for option in stage_options:
                if option == 'sbatch_options':
                    self.sbatch_options.update(stage_options[option])
                else:
                    self.logger.debug('ignoring option: '+option)
        return RemoteHost.dispatch(self, **kwargs)
    def filter_script_content(self, content):
        return RemoteHost.filter_script_content(self,
                Slurm.filter_script_content(self, content))
    def make_job(self, stage_index=None, source=None, region_index=None, segment_index=None):
        Slurm.make_job(self, stage_index, self.format_source(source),
                region_index, segment_index)
    def submit_jobs(self):
        sbatch_script = self.make_sbatch_script()
        dest = '/'.join((self.wd, os.path.basename(sbatch_script)))
        attrs = self.ssh.put(sbatch_script, dest)
        self.logger.debug(attrs)
        self.logger.info('sbatch script transferred to: '+dest)
        #self.logger.info('running: module load singularity')
        #out, err = self.ssh.exec('module load singularity')
        if self.remote_dependencies():
            cmd = self.remote_dependencies()+'; sbatch '+dest
        else:
            cmd = 'sbatch '+dest
        self.logger.info('running: '+cmd)
        out, err = self.ssh.exec(cmd, shell=True)
        if out:
            out = out.rstrip()
            self.logger.debug(out)
            if out.startswith('Submitted batch job '):
                self.job_id = out.split(' ')[-1]
        if err:
            self.logger.error(err.rstrip())
        self.pending_jobs = []
    def wait_for_job_completion(self):
        self.logger.info('''\
notice: job failures are not reported before the stage is complete;
        check the .err log files generated in the remote working directory
        and manually interrupt the pipeline if all the jobs seem to fail\
''')
        total_task_count = None
        try:
            cmd = 'squeue -j {} -h -o "%K %t %M %R"'.format(self.job_id)
            while True:
                time.sleep(self.refresh_interval)
                out, err = self.ssh.exec(cmd, shell=True)
                if err:
                    err = err.rstrip()
                    if err == 'slurm_load_jobs error: Invalid job id specified':
                        # complete
                        break
                    self.logger.error(err.rstrip())
                elif out:
                    # parse and print progress info
                    start, _out = None, []
                    _continue = False
                    for line in out.splitlines():
                        parts = line.split()
                        if '-' in parts[0]:
                            if start is None:
                                try:
                                    start, stop = parts[0].split('-')
                                except ValueError:
                                    self.logger.debug('squeue output parsing failed: \n'+out)
                                    _continue = True; break
                                else:
                                    stop = stop.split('%')[0]
                                    start, stop = int(start), int(stop)
                            else:
                                self.logger.debug('squeue output parsing failed: \n'+out)
                                _continue = True; break
                        elif '%' in parts[0]:
                            start = stop = int(parts[0].split('%')[0])
                        else:
                           _out.append(parts)
                    if _continue:
                        continue
                    if start is None:
                        total, pending = total_task_count, 0
                    else:
                        total = stop + 1
                        pending = stop - start + 1

                        if total_task_count is None:
                            total_task_count = total
                        elif total_task_count != total:
                            self.logger.debug('could not estimate the total number of tasks from squeue output')
                            total_task_count = total

                    running = 0
                    other = 0
                    for out in _out:
                        try:
                            array_ix, status, time_used = int(out[0]), out[1], out[2]
                        except ValueError:
                            self.logger.debug('squeue output parsing failed on line: '+' '.join(out))
                            continue
                        reason = ' '.join(out[3:])
                        if status == 'R':
                            running += 1
                        else:
                            other += 1
                        #self.logger.debug(task: {:d}   status: {}   time used: {}   reason: {}'.format(array_ix, status, time_used, reason))
                    self.logger.info('tasks:\t{} done,\t{} running,\t{} pending{}'.format(
                        '(unknown)' if total is None else total-pending-running-other,
                        '(unknown)' if running is None else running,
                        '(unknown)' if pending is None else pending,
                        ',\t{} in abnormal state'.format(other) if other else ''))
                else:
                    # complete
                    break
        except:
            self.logger.info('killing jobs with: scancel '+self.job_id)
            self.ssh.exec('scancel '+self.job_id, shell=True)
            raise
    @property
    def collection_interpreter(self):
        if self._srun_options:
            cmd = ['srun']
            for option, value in self.srun_options.items():
                if option[1:]:
                    fmt = '--{}={}'
                else:
                    fmt = '-{} {}'
                if isinstance(value, str) and ' ' in value:
                    value = '"{}"'.format(value)
                cmd.append(fmt.format(option, value))
            cmd.append(self.interpreter)
            return ' '.join(cmd)
        else:
            return self.interpreter
    def collect_results(self, stage_index=None, inplace=False, reload_existing_rwa_files=False):
        RemoteHost.collect_results(self, '*.out', stage_index, inplace=inplace,
                reload_existing_rwa_files=reload_existing_rwa_files)
    collect_results.__doc__ = RemoteHost.collect_results.__doc__
    def delete_temporary_data(self):
        RemoteHost.delete_temporary_data(self)
        Slurm.delete_temporary_data(self)
    def resume(self, log=None, wd=None, stage_index=None, job_id=None):
        """
        Parses log output of the disconnected instance, looks for the current stage index,
        and tries to collect the resulting files.

        This completes the current stage only. Further stages are not run.

        Arguments:

            log (str): log output; progress information can be omitted

            wd (str): location of the working directory on the remote host

            stage_index (*int* or *list*): index of the stage(s) to resume

            job_id (str): Slurm job array ID

        """
        if wd is None or stage_index is None or job_id is None:
            if log is None:
                log = input('please copy-paste below the log output of the disconnected instance\n(job progress information can be omitted):\n')
            log = log.splitlines()
            for line in log[::-1]:
                if wd:
                    try:
                        opt = line.index(' --stage-index=')
                    except ValueError:
                        pass
                    else:
                        stage_index = line[opt+15:].split()[0]
                        stage_index = [ int(s) for s in stage_index.split(',') ]
                        break
                elif job_id:
                    assert line.startswith('running: ') and 'sbatch ' in line
                    script = line.split('sbatch ')[-1].rstrip()
                    if not wd:
                        wd = '/'.join(script.split('/')[:-1])
                elif line.startswith('Submitted batch job '):
                    job_id = line[20:].rstrip()
                elif line.startswith('working directory: '):
                    wd = line[19:].rstrip()
            if stage_index is None and len(self._eldest_parent.pipeline)==1:
                stage_index = 0
        if stage_index is None:
            self.logger.info('cannot identify an execution point where to resume from')
        else:
            self.setup(sys.executable, create_working_directory=False)
            assert self.submit_side
            #self.delete_temporary_data() # undo wd creation during setup
            self.working_directory = wd
            self.job_id = job_id
            if isinstance(stage_index, int):
                stage_index = [stage_index]
            self.logger.info('trying to complete stage(s):\t'+', '.join([str(i) for i in stage_index])+"""
 with working directory:\t'{}'
  and job id:\t\t\t{}\
""".format(wd, job_id))
            if self.job_id is not None:
                self.wait_for_job_completion()
            stage = self._eldest_parent.pipeline._stage[stage_index[-1]]
            self.collect_results(stage_index=stage_index,
                    reload_existing_rwa_files=stage.update_existing_rwa_files)

    def pack_temporary_rwa_files(self, log=None, wd=None, stage_index=None, job_id=None):
        """
        Similarly to :meth:`resume`, `pack_temporary_rwa_files` combines the intermediate
        *.rwa* files in the temporary directory for a running instance on the compute host.

        This does not move files out of the temporary directory.
        Instead, the intermediate *.rwa* files are replaced by new (but fewer) intermediate
        *.rwa* files that result from combining the former files.

        This may be useful to save disk space, as the localization data is duplicated as
        many times as there are intermediate files.

        .. warning::

            not tested yet

        """
        # perform the same checks as `resume`, although the temporary directory only is
        # required
        if wd is None or stage_index is None or job_id is None:
            if log is None:
                log = input('please copy-paste below the log output of the disconnected instance\n(job progress information can be omitted):\n')
            log = log.splitlines()
            for line in log[::-1]:
                if wd:
                    try:
                        opt = line.index(' --stage-index=')
                    except ValueError:
                        pass
                    else:
                        stage_index = line[opt+15:].split()[0]
                        stage_index = [ int(s) for s in stage_index.split(',') ]
                        break
                elif job_id:
                    assert line.startswith('running: ') and 'sbatch ' in line
                    script = line.split('sbatch ')[-1].rstrip()
                    if not wd:
                        wd = '/'.join(script.split('/')[:-1])
                elif line.startswith('Submitted batch job '):
                    job_id = line[20:].rstrip()
                elif line.startswith('working directory: '):
                    wd = line[19:].rstrip()
            if stage_index is None and len(self._eldest_parent.pipeline)==1:
                stage_index = 0
        if stage_index is None:
            self.logger.info('cannot identify an execution point where to resume from')
        else:
            self.setup(sys.executable, create_working_directory=False)
            assert self.submit_side
            #self.delete_temporary_data() # undo wd creation during setup
            self.working_directory = wd
            self.job_id = job_id
            if isinstance(stage_index, int):
                stage_index = [stage_index]
            self.logger.info('packing intermediate files for uncomplete stage(s):\t'+', '.join([str(i) for i in stage_index])+"""
 with working directory:\t'{}'
  and job id:\t\t\t{}\
""".format(wd, job_id))
            self.collect_results(stage_index=stage_index, inplace=True)
    def list_uncomplete_task_log_files(self):
        cmd = "for file in `ls '{}/*.err'`; do if [ -z \"`tail -n1 '$file' | grep -Ex 'stage [0-9]{1,} done'`\" ]; then echo $file; fi; done".format(self.wd)
        self.logger.debug('running: '+cmd)
        out, err = self.ssh.exec(cmd, shell=True)
        if err:
            self.logger.error(err.rstrip())
        if out:
            out = out.rstrip()
            return out.splitlines()
        else:
            return []


Environment.register(SlurmOverSSH)


class SingularitySlurm(SlurmOverSSH):
    """
    Runs TRamWAy jobs as Slurm jobs in a Singularity container.

    The current default Singularity container is *tramway-hpc-210722.sif*.
    See also `available_images.rst <https://github.com/DecBayComp/TRamWAy/blob/master/containers/available_images.rst>`_.

    Children classes should define the :meth:`hostname` and :meth:`scratch` methods.
    They can be defined as standard methods or class methods.
    """
    @classmethod
    def hostname(cls):
        raise NotImplementedError
    @classmethod
    def scratch(cls, username):
        raise NotImplementedError
    @classmethod
    def singularity_options(cls):
        return ''
    def __init__(self, **kwargs):
        SlurmOverSSH.__init__(self, **kwargs)
        singularity_options = self.singularity_options()
        if singularity_options and singularity_options[0] != ' ':
            singularity_options = ' '+singularity_options
        default_container = self.default_container()
        self.interpreter = f'singularity exec -H $HOME{singularity_options} {default_container} python3.6 -s'
        self.ssh.host = self.hostname()
    @property
    def username(self):
        try:
            username, _ = self.ssh.host.split('@')
        except (AttributeError, ValueError):
            import getpass
            username = getpass.getuser()
        return username
    @username.setter
    def username(self, name):
        if name is None:
            self.ssh.host = self.hostname()
            self.wd = None
        else:
            self.ssh.host = '@'.join((name, self.hostname()))
            if self.wd is None:
                self.wd = self.scratch(name)
    @property
    def container(self):
        parts = self.interpreter.split()
        return parts[parts.index('python3.6')-1]
    @container.setter
    def container(self, path):
        parts = self.interpreter.split()
        p = parts.index('python3.6')
        self.interpreter = ' '.join(parts[:p-1]+[path]+parts[p:])
    def get_container_url(self, container=None):
        if container is None:
            container = self.container
        return {
                'tramway-hpc-200928.sif':   'http://dl.pasteur.fr/fop/0G0Tnjd8/tramway-hpc-200928.sif',
                'tramway-hpc-210112.sif':   'http://dl.pasteur.fr/fop/tVZe8prV/tramway-hpc-210112.sif',
                'tramway-hpc-210114.sif':   'http://dl.pasteur.fr/fop/cZWZqsDW/tramway-hpc-210114.sif',
                'tramway-hpc-210125.sif':   'http://dl.pasteur.fr/fop/6Avu9HuV/tramway-hpc-210125.sif',
                'tramway-hpc-210201.sif':   'http://dl.pasteur.fr/fop/MSRwa8CR/tramway-hpc-210201.sif',
                'tramway-hpc-210222.sif':   'http://dl.pasteur.fr/fop/rzx2LnjB/tramway-hpc-210222.sif',
                'tramway-hpc-210301.sif':   'http://dl.pasteur.fr/fop/7jSSZCnb/tramway-hpc-210301.sif',
                'tramway-hpc-210302.sif':   'http://dl.pasteur.fr/fop/53bfSkmM/tramway-hpc-210302.sif',
                'tramway-hpc-210330.sif':   'http://dl.pasteur.fr/fop/FrDZQBuy/tramway-hpc-210330.sif',
                'tramway-hpc-210527.sif':   'http://dl.pasteur.fr/fop/7F0LaOEX/tramway-hpc-210527.sif',
                'tramway-hpc-210608.sif':   'http://dl.pasteur.fr/fop/5LCsGe80/tramway-hpc-210608.sif',
                'tramway-hpc-210628.sif':   'http://dl.pasteur.fr/fop/Cr969IPb/tramway-hpc-210628.sif',
                'tramway-hpc-210715.sif':   'http://dl.pasteur.fr/fop/lzFiPalM/tramway-hpc-210715.sif',
                'tramway-hpc-210720.sif':   'http://dl.pasteur.fr/fop/rb4blYsf/tramway-hpc-210720.sif',
                'tramway-hpc-210722.sif':   'http://dl.pasteur.fr/fop/kxCZAcy0/tramway-hpc-210722.sif',
                }.get(container, None)
    @classmethod
    def default_container(cls):
        return 'tramway-hpc-210722.sif'
    def setup(self, *argv, **kwargs):
        SlurmOverSSH.setup(self, *argv, **kwargs)
        if self.submit_side:
            self.ssh.download_if_missing(self.container, self.get_container_url(), self.logger)
    @property
    def working_directory(self):
        if self._working_directory is None:
            self._working_directory = self.scratch(self.username)
        return self._working_directory
    working_directory.__doc__ = SlurmOverSSH.working_directory.__doc__
    @working_directory.setter
    def working_directory(self, wd):
        self._working_directory = wd


class Tars(SingularitySlurm):
    """
    Designed for server *tars.pasteur.fr*.

    The server is closed.
    """
    @classmethod
    def hostname(cls):
        return 'tars.pasteur.fr'
    @classmethod
    def scratch(cls, username):
        return '/'.join(('/pasteur/scratch/users', username))
    @classmethod
    def singularity_options(cls):
        return ' -B /pasteur'
    @classmethod
    def remote_dependencies(cls):
        return 'module load singularity'


class GPULab(SingularitySlurm):
    """
    Designed for server *adm.inception.hubbioit.pasteur.fr*.
    """
    @classmethod
    def hostname(cls):
        return 'adm.inception.hubbioit.pasteur.fr'
    @classmethod
    def scratch(cls, username):
        return '/'.join(('/master/home', username, 'scratch'))
    def __init__(self, **kwargs):
        SingularitySlurm.__init__(self, **kwargs)
        self.interpreter = self.interpreter.replace('-B /pasteur ', '')


class Maestro(SingularitySlurm):
    """
    Designed for server *maestro.pasteur.fr*.
    """
    @classmethod
    def hostname(cls):
        return 'maestro.pasteur.fr'
    @classmethod
    def scratch(cls, username):
        return '/'.join(('/pasteur/sonic/scratch/users', username))
    @classmethod
    def singularity_options(cls):
        return ' -B /pasteur'
    @classmethod
    def remote_dependencies(cls):
        return 'module load singularity'


__all__ = ['Environment', 'LocalHost', 'SlurmOverSSH', 'Tars', 'GPULab', 'Maestro']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..artefact import analysis, Analysis
from .abc import Time
import tramway.tessellation.time as time
import tramway.tessellation.window as window
import numpy as np
import pandas as pd


class DT(object):
    """
    Implements the default behavior of methods common to the initializer and the
    specialized attributes.

    It gives access to the frame interval (the :attr:`dt` and :attr:`time_step`
    attributes are aliases of :attr:`frame_interval`),
    and features the :meth:`as_time_segments` slicer.

    The default implementation suits the initializer's behavior,
    i.e. a single all-in time segment.
    """
    __slots__ = ()
    def enable_regularization(self):
        """
        This method is called before running the inference plugins that define
        *time_prior* parameters.
        """
        raise AttributeError('no time segmentation defined; cannot regularize in time')
    def n_time_segments(self, sampling):
        """
        Returns the number of time segments (*int*) that `sampling` includes,
        under the hypothesis that `sampling` was generated following
        this segments definition (*self*).
        """
        return 1
    def as_time_segments(self, sampling, maps=None, index=None, return_index=False, return_times=True):
        if not (index is None or index == 0):
            raise ValueError('no time segments defined')
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        if return_times:
            try:
                t = sampling.points['t']
                all_times = np.r_[t.min(), t.max()]
            except (AttributeError, KeyError):
                all_times = None
        if maps is None:
            if return_times:
                if return_index:
                    yield 0, all_times, sampling
                else:
                    yield all_times, sampling
            else:
                if return_index:
                    yield 0, sampling
                else:
                    yield sampling
        else:
            if return_times:
                if return_index:
                    yield 0, all_times, sampling, maps
                else:
                    yield all_times, sampling, maps
            else:
                if return_index:
                    yield 0, sampling, maps
                else:
                    yield sampling, maps
    as_time_segments.__doc__ = Time.as_time_segments.__doc__
    @property
    def spt_data(self):
        return self._parent.spt_data
    @property
    def dt(self):
        """
        *float*: See :attr:`~tramway.analyzer.spt_data.SPTParameters.dt`
        """
        return self.spt_data.dt
    @dt.setter
    def dt(self, dt):
        self.spt_data.dt = dt
    @property
    def time_step(self):
        """
        *float*: See :attr:`~tramway.analyzer.spt_data.SPTParameters.time_step`
        """
        return self.spt_data.time_step
    @time_step.setter
    def time_step(self, dt):
        self.spt_data.time_step = dt
    @property
    def frame_interval(self):
        """
        *float*: See :attr:`~tramway.analyzer.spt_data.SPTParameters.frame_interval`
        """
        return self.spt_data.frame_interval
    @frame_interval.setter
    def frame_interval(self, dt):
        self.spt_data.frame_interval = dt
    def get_spatial_segmentation(self, sampling, segment=None):
        """
        Returns the spatial tessellation.
        """
        if not (segment is None or segment == 0):
            raise ValueError('multiple time segments expected, but no time segmentation defined')
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        mesh = sampling.tessellation
        if isinstance(mesh, time.TimeLattice):
            raise TypeError('time segments found, but no time segmentation defined')
        return mesh

class TimeInitializer(Initializer, DT):
    """
    Initializer :class:`Time` class for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.time` main attribute.

    The :attr:`~tramway.analyzer.RWAnalyzer.time` attribute self-modifies on
    calling *from_...* methods.
    """
    __slots__ = ()
    def __init__(self, *args, **kwargs):
        Initializer.__init__(self, *args, **kwargs)
    def from_sliding_window(self, duration=None, shift=None):
        """
        Defines a sliding time window.

        Arguments:

            duration (float): window duration in seconds

            shift (float): time shift between successive segments, in seconds;
                by default, equals to `duration`

        See also :class:`SlidingWindow`.
        """
        self.specialize( SlidingWindow, duration, shift )
    def from_sampling(self, sampling=None, verbose=False):
        """
        Extracts the time segmentation parameters stored in a
        :class:`~tramway.tessellation.base.Partition` object
        and tries to initialize the parent
        :attr:`~tramway.analyzer.RWAnalyzer.time` attribute
        correspondingly.

        This may fail,
        either silently or raising a :class:`ValueError` exception,
        as this method covers only cases of sliding windows.

        If argument `sampling` is not an `Analysis` or `Partition`
        object, it is taken as the label of an artefact in any
        analysis tree.
        """
        if sampling is None or isinstance(sampling, (int, str)) or callable(sampling):
            label = sampling
            sampling = first(self._eldest_parent.roi.as_support_regions()).get_sampling(label)
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        segmentation = sampling.tessellation
        if isinstance(segmentation, window.SlidingWindow):
            duration, shift, start_time = segmentation.duration, segmentation.shift, segmentation.start_time
        else:
            # experimental!
            if isinstance(segmentation, time.TimeLattice):
                segments = segmentation.time_lattice
                if segments is None or len(segments)==0:
                    raise ValueError('no time segments defined')
                durations = np.diff(segments, axis=1)
                duration = durations[0]
                if not np.all(durations == duration):
                    raise ValueError('varying time segment duration is not supported')
                shifts = np.diff(segments[:,0])
                if shifts.size == 0:
                    shift = None
                else:
                    shift = shifts[0]
                    if not np.all(shifts == shift):
                        raise ValueError('varying time shift is not supported')
                start_time = segments[0,0]
                t0 = sampling.points['t'].min()
                if start_time == t0:
                    start_time = None
            else:
                self.logger.warning('unsupported time segmentation type: '+str(type(segmentation)))
                return
        self.specialize( SlidingWindow, duration, shift )
        if start_time is not None:
            self._parent.time.start_time = start_time
        if verbose:
            logger = self.logger
            prm = [('duration', duration)]
            if shift is not None:
                prm.append(('shift', shift))
            if start_time is not None:
                prm.append(('start_time', start_time))
            msg = "loading sliding window with parameters:\n - " + \
                    "\n - ".join([ "{}: {}".format(key, val) for key, val in prm ])
            logger.info(msg)


class SlidingWindow(AnalyzerNode, DT):
    """
    Specialization for the :attr:`~tramway.analyzer.RWAnalyzer.time` attribute.

    Defines the :meth:`segment` method that segments the SPT data into time segments
    and combines with the spatial tessellation.
    """
    __slots__ = ('_duration', '_shift', '_start_time', '_regularize_in_time')
    @property
    def reified(self):
        return True
    def __init__(self, duration=None, shift=None, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._duration = None if duration is None else float(duration)
        self._shift = None if shift is None else float(shift)
        self._start_time = None
        self._regularize_in_time = False
    @property
    def duration(self):
        """ *float*: Window duration in seconds """
        return self._duration
    @duration.setter
    def duration(self, d):
        self._duration = None if d is None else float(d)
    @property
    def shift(self):
        """ *float*: Time shift between successive time segments, in seconds """
        return self.duration if self._shift is None else self._shift
    @shift.setter
    def shift(self, s):
        self._shift = None if s is None else float(s)
    @property
    def window_duration(self):
        """ *float*: Alias for :attr:`duration` """
        return self.duration
    @window_duration.setter
    def window_duration(self, d):
        self.duration = d
    @property
    def window_shift(self):
        """ *float*: Alias for :attr:`shift` """
        return self.shift
    @window_shift.setter
    def window_shift(self, s):
        self.shift = s
    @property
    def start_time(self):
        """
        *float*: Start time for running the sliding window.
            By default, the window starts from the first data point in the input data
            (usually the ROI-cropped data, NOT the entire SPT dataset).
        """
        if self._start_time == 'sync':
            return self.spt_data.bounds.loc['min','t']
        return self._start_time
    @start_time.setter
    def start_time(self, t0):
        self._start_time = float(t0) if isinstance(t0, int) else t0
    def sync_start_times(self):
        """ aligns the start times of all the ROI to the same minimum time.

        Beware this may load all the SPT data files.
        Setting the :attr:`~tramway.analyzer.spt_data.SPTData.bounds` attribute
        of the :attr:`~tramway.analyzer.RWAnalyzer.spt_data` main attribute
        discards the need for screening the SPT data."""
        self._start_time = self.spt_data.bounds.loc['min','t']
    @property
    def time_window_kwargs(self):
        kwargs = dict(duration=self.duration, shift=self.shift, start_time=self.start_time)
        if self.regularize_in_time:
            kwargs['time_dimension'] = True
        return kwargs
    @analysis
    def segment(self, spt_dataframe, tessellation=None):
        tess = window.SlidingWindow(**self.time_window_kwargs)
        tess.spatial_mesh = tessellation
        try:
            spt_dataframe = spt_dataframe[self.tesseller.colnames]
        except AttributeError:
            pass
        tess.tessellate(spt_dataframe, time_only=True)
        return tess
    segment.__doc__ = Time.segment.__doc__
    @property
    def tesseller(self):
        return self._parent.tesseller
    def enable_regularization(self):
        self._regularize_in_time = True
    @property
    def regularize_in_time(self):
        """ *bool*: :const:`True` if time regularization is enabled """
        return self._regularize_in_time
    def get_time_segments(self, sampling):
        """
        Returns the time segment bounds as an Nx2 array.
        """
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        return sampling.tessellation.time_lattice
    def get_spatial_segmentation(self, sampling, segment=None):
        """
        Returns the spatial tessellation.
        """
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        return sampling.tessellation.spatial_mesh
    def n_time_segments(self, sampling):
        return len(self.get_time_segments(sampling))
    def as_time_segments(self, sampling, maps=None, index=None, return_index=False, return_times=True):
        if return_index:
            if return_times or maps is not None:
                def _indexer(*args):
                    for i, res in indexer(*args, return_index=True):
                        yield (i,)+res
            else:
                def _indexer(*args):
                    yield from indexer(*args, return_index=True)
        else:
            _indexer = indexer
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        it = sampling.tessellation.split_segments(sampling, return_times=return_times)
        if maps is not None:
            if isinstance(maps, Analysis):
                maps = maps.data
            try:
                maps = maps.maps
            except AttributeError:
                pass
            maps = sampling.tessellation.split_segments(maps)
            if return_times:
                ts, partitions = zip(*it)
                it = zip(ts, partitions, maps)
            else:
                partitions = it
                it = zip(partitions, maps)
        yield from _indexer(index, it)
    @property
    def spt_data(self):
        return self._parent.spt_data
    def self_update(self, op):
        self._parent._time = op(self)
    def segment_label(self, map_label, times, sampling):
        """
        Makes a label combining the input label as prefix and time-related suffix.
        """
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        if isinstance(times, int):
            segment_index = times
            times = sampling.tessellation.time_lattice[segment_index]
        format_time = lambda t: '{:.3f}'.format(t).rstrip('0').rstrip('.')
        if map_label:
            return '{} -- t={}-{}s'.format(map_label, *[ format_time(t) for t in times ])
        else:
            return 't={}-{}s'.format(*[ format_time(t) for t in times ])
    def combine_segments(self, combined_output_label, combined_sampling, commit=True, permissive=False):
        analyses = combined_sampling.subtree
        sampling = combined_sampling.data
        try:
            single_segment_output, labels = [], []
            for times in sampling.tessellation.time_lattice:
                single_segment_label = self.segment_label(combined_output_label, times, sampling)
                labels.append(single_segment_label)
                single_segment_output.append(analyses[single_segment_label].data)
        except KeyError:
            if single_segment_output:
                if permissive:
                    self.logger.warning('not all segments are available; combining aborted')
                else:
                    raise KeyError('not all segments are available; combining aborted') from None
            else:
                if permissive:
                    return None
                else:
                    labels = list(analyses.labels)
                    if labels and any( l.startswith(combined_output_label+' -- t=') for l in labels ):
                        raise KeyError("'{}' segments do not start at the expected time; expected label: '{}'; found labels: {}".format(combined_output_label, single_segment_label, str(labels))) from None
                    else:
                        raise KeyError("no '{}' segments found".format(combined_output_label)) from None
        assert single_segment_output
        from tramway.inference.base import Maps
        ok = [ isinstance(m, Maps) for m in single_segment_output ]
        if all(ok):
            pass
        elif any(ok):
            raise TypeError('cannot combine heterogeneous types')
        else:
            raise TypeError('cannot combine values of type: {}'.format(type(single_segment_output[0])))
        ncells = sampling.tessellation.spatial_mesh.number_of_cells
        import copy
        it = iter(single_segment_output)
        maps = copy.deepcopy(next(it))
        maps.runtime = None
        maps.posteriors = None
        df, s = [maps.maps], 0
        while True:
            try:
                m = next(it)
            except StopIteration:
                break
            s += 1
            m = m.maps.copy()
            m.index += s*ncells
            df.append(m)
        maps.maps = pd.concat(df)
        if commit:
            for label in labels:
                del analyses[label]
            analyses[combined_output_label] = maps
        return maps

Time.register(SlidingWindow)


__all__ = ['Time', 'DT', 'TimeInitializer', 'SlidingWindow']




#!/usr/bin/env python

from tramway.core import *
from tramway.core.hdf5 import *
from tramway.tessellation import CellStats
from tramway.inference import Distributed, Translocations, Maps, distributed, neighbours_per_axis

import numpy as np
import pandas as pd
import scipy.sparse as sparse
from collections import namedtuple, OrderedDict

import os
import argparse
try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser


# cluster
__zone__ = 'ZONE:'
__ntr__ = 'NUMBER_OF_TRANSLOCATIONS:'
__x__ = 'X-CENTRE:'
__y__ = 'Y-CENTRE:'
__area__ = 'AREA:'
__convhull__ = 'AREA_CONVHULL:'
__nleft__ = 'NUMBER_OF_LEFT_NEIGHBOURS:'
__nright__ = 'NUMBER_OF_RIGHT_NEIGHBOURS:'
__ntop__ = 'NUMBER_OF_TOP_NEIGHBOURS:'
__nbottom__ = 'NUMBER_OF_BOTTOM_NEIGHBOURS:'
__left__ = 'LEFT_NEIGHBOURS:'
__right__ = 'RIGHT_NEIGHBOURS:'
__top__ = 'TOP_NEIGHBOURS:'
__bottom__ = 'BOTTOM_NEIGHBOURS:'
__coord__ = ['DX', 'DY', 'DT']


def import_cluster_file(input_file, return_areas=False):
    with open(input_file, 'r') as f:
        lines = f.readlines()
    if return_areas:
        areas = []
    cells = OrderedDict()
    ncells = 0
    adjacency = []
    i = None
    ntr_defined = x_defined = y_defined = area_defined = convhull_defined = False
    C = OrderedDict
    neighbours_prefixes, neighbours, nneighbours, neighbours_code = C(), C(), C(), C()
    for side, prefixes, code in (
            ('left', (__left__, __nleft__), -1),
            ('right', (__right__, __nright__), 1),
            ('top', (__top__, __ntop__), 2),
            ('bottom', (__bottom__, __nbottom__), -2),
        ):
        neighbours_prefixes[side] = prefixes # static
        neighbours[side] = [] # = undefined
        nneighbours[side] = None # = undefined
        neighbours_code[side] = code # static
    coord_section = False
    for l, line in enumerate(lines+['ZONE: -1']):
        line = line.strip()
        if not line or line[0] in '#;':
            continue
        if line.split() == __coord__:
            if i is None:
                raise IOError('not a cluster file')
            coord_section = True
            continue
        elif coord_section:
            try:
                tr = [ float(k) for k in line.split() ]
            except ValueError:
                if ntr_defined and len(coord) != ntr:
                    print('cell {}: wrong number of translocations'.format(i))
                coord_section = False
            else:
                coord.append(tr)
                continue
        if line.startswith(__zone__):
            if i is not None:
                # finalize current cell
                cols = ['x', 'y', 't']
                cell = Translocations(i - 1, pd.DataFrame(np.array(coord), columns=cols))
                cell.space_cols = cols[:2]
                cell.time_col = cols[-1]
                if x_defined and y_defined:
                    cell.center = np.array([x, y])
                if area_defined and return_areas:
                    areas.append((i - 1, area))
                if convhull_defined:
                    cell.volume = convhull
                cells[i - 1] = cell
                ncells = max(i, ncells)
                all_neighbours = []
                for side in neighbours:
                    if neighbours[side]:
                        if not (nneighbours[side] is None or len(neighbours[side]) == nneighbours[side]):
                            print('cell {}: wrong number of {} neighbours'.format(i, side))
                        all_neighbours += [ (j, neighbours_code[side]) for j in neighbours[side] ]
                    elif 0 < nneighbours[side]:
                        print('cell {}: {} neighbours are not defined'.format(i, side))
                if all_neighbours:
                    ncells = max(ncells, *[ j for j,_ in all_neighbours ])
                    adjacency += [ (i-1,j-1,k) for j,k in all_neighbours ]
            # initialize new cell parsing
            i = int(line[len(__zone__):])
            ntr_defined = x_defined = y_defined = area_defined = convhull_defined = False
            for side in neighbours:
                neighbours[side], nneighbours[side] = [], None
            coord = []
        elif line.startswith(__ntr__):
            ntr = int(line[len(__ntr__):])
            ntr_defined = True
        elif line.startswith(__x__):
            x = float(line[len(__x__):])
            x_defined = True
        elif line.startswith(__y__):
            y = float(line[len(__y__):])
            y_defined = True
        elif line.startswith(__area__):
            area = float(line[len(__area__):])
            area_defined = True
        elif line.startswith(__convhull__):
            convhull = float(line[len(__convhull__):])
            convhull_defined = True
        else:
            cont = False
            for side in neighbours_prefixes:
                prefix, prefix_n = neighbours_prefixes[side]
                if line.startswith(prefix_n):
                    if nneighbours[side] is not None:
                        print('cell {}: duplicate field for the number of {} neighbours'.format(i, side))
                    try:
                        nneighbours[side] = int(line[len(prefix_n):])
                    except ValueError:
                        print('cell {}: wrong format for the number of {} neighbours'.format(i, side))
                    else:
                        cont = True
                        break
                elif line.startswith(prefix):
                    if neighbours[side]:
                        print('cell {}: duplicate field for {} neighbours'.format(i, side))
                    try:
                        neighbours[side] = [ int(k) for k in line[len(prefix):].split() ]
                    except ValueError:
                        print('cell {}: wrong format for {} neighbours'.format(i, side))
                    else:
                        cont = True
                        break
            if not cont:
                print('skipping line {}:'.format(l))
                print(line)
    # make the Distributed object
    i, j, k = zip(*adjacency)
    adjacency = sparse.coo_matrix((k, (i, j)), shape=(ncells, ncells)).tocsr()
    for i in cells:
        cell = cells[i]
        neighbours = np.vstack([ cells[j].center for j in adjacency[i].indices ])
        cell.span = neighbours - cell.center[np.newaxis,:]
    distr = Distributed(cells, adjacency)

    if return_areas:
        index, areas = zip(*areas)
        areas = pd.Series(areas, index=index)
        return distr, areas
    else:
        return distr


def import_vmesh_file(input_file):
    maps = pd.read_table(input_file, skiprows=7, index_col=0)
    maps.index -= 1
    maps = maps.rename(columns={'x-Force': 'force x', 'y-Force': 'force y',
        'Translocations': 'n', 'x-Centre': 'x', 'y-Centre': 'y',
        'Diffusion': 'diffusivity', 'Potential': 'potential'})
    maps = Maps(maps)
    with open(input_file, 'r') as f:
        while True:
            line = f.readline().rstrip()
            if not line:
                break
            key, value = line.split(': ')
            if key == 'Input File':
                pass
            elif key == 'Inference Mode':
                maps.mode = value
            elif key == 'Localization Error':
                value, unit = value.split()
                assert unit == '[nm]'
                value = float(value) * 1e3
                maps.localization_error = value
            elif key == 'Number of Zones':
                pass
            elif key == 'Diffusion Prior':
                maps.diffusivity_prior = float(value)
            elif key == 'Potential Prior':
                maps.potential_prior = float(value)
            elif key == "Jeffreys' Prior":
                maps.jeffreys_prior = bool(value)
    return maps


def import_file(root_file, label, cluster_file=None, vmesh_file=None, output_file=None, **kwargs):
    clabel = vlabel = None
    if isinstance(label, (tuple, list)):
        if label:
            if label[1:]:
                try:
                    clabel, vlabel = label
                except ValueError:
                    raise ValueError('too many labels')
            else:
                clabel = label[0]
    else:
        clabel = label # can be None
    try:
        tree = load_rwa(root_file, verbose=False)
    except:
        trajectory_file = root_file
        xyt = load_xyt(trajectory_file)
        tree = Analyses(xyt)
        if not output_file:
            output_file = os.path.splitext(root_file)[0] + '.rwa'
    else:
        if not output_file:
            output_file = root_file
    if cluster_file:
        distr = import_cluster_file(cluster_file)
        clabel = tree.autoindex(clabel)
        tree.add(distr, label=clabel)
    if vmesh_file:
        maps = import_vmesh_file(vmesh_file)
        tree[clabel].add(maps, label=vlabel)
    save_rwa(output_file, tree, force=True, **kwargs)


def export_to_cluster_file(cells, cluster_file, neighbours=None, neighbours_kwargs={}, new_cell=Translocations, distributed_kwargs={}, **kwargs):
    not_2D = ValueError('data are not 2D')
    if isinstance(cells, Distributed):
        if cells.dim != 2:
            raise not_2D
        zones = cells
    else:
        if cells.tessellation._cell_centers.shape[1] != 2:
            raise not_2D
        zones = distributed(cells, new_cell=new_cell, **distributed_kwargs)
    if neighbours is None:
        neighbours = neighbours_per_axis
    with open(cluster_file, 'w') as f:
        for i in zones:
            zone = zones[i]
            f.write('ZONE: {}\n'.format(i+1))
            f.write('NUMBER_OF_TRANSLOCATIONS: {}\n'.format(len(zone)))
            f.write('X-CENTRE: {}\n'.format(zone.center[0]))
            f.write('Y-CENTRE: {}\n'.format(zone.center[1]))
            #area = np.prod(np.max(zone.span, axis=0) - np.min(zone.span, axis=0))
            area = zone.volume
            f.write('AREA: {}\n'.format(area))
            f.write('AREA_CONVHULL: {}\n'.format(area))
            below, above = neighbours(i, zones, **neighbours_kwargs)
            _neighbours = zones.adjacency[i].indices
            _neighbours = OrderedDict(
                left = _neighbours[below[0]],
                right = _neighbours[above[0]],
                top = _neighbours[above[1]],
                bottom = _neighbours[below[1]],
                )
            for side in _neighbours:
                f.write('NUMBER_OF_{}_NEIGHBOURS: {}\n'.format(
                    side.upper(), _neighbours[side].size))
                if _neighbours[side].size:
                    f.write('{}_NEIGHBOURS: '.format(side.upper()))
                    for j in _neighbours[side]:
                        f.write('{}\t'.format(j+1))
                    f.write('\n')
            f.write('DX\tDY\tDT\n')
            for dx, dy, dt in np.hstack((zone.dr, zone.dt[:,np.newaxis])):
                f.write('{}\t{}\t{}\n'.format(dx, dy, dt))
            f.write('\n\n')


def export_to_vmesh_file(cells, maps, vmesh_file, cluster_file=None, auto=False, new_cell=Translocations, distributed_kwargs={}, **kwargs):
    not_2D = ValueError('data are not 2D')
    if isinstance(cells, Distributed):
        if cells.dim != 2:
            raise not_2D
        zones = cells
    else:
        if cells.tessellation._cell_centers.shape[1] != 2:
            raise not_2D
        zones = distributed(cells, new_cell=new_cell, **distributed_kwargs)
    with open(vmesh_file, 'w') as f:
        # Input File
        if cluster_file is None and auto:
            dirname, basename = os.path.split(vmesh_file)
            cluster_files = [ fn
                for fn in os.listdir(dirname if dirname else '.')
                if fn.endswith('.cluster') and basename.startswith(fn[:8]) ]
            if cluster_files:
                if cluster_files[1:]:
                    print('multiple cluster files match:')
                    for fn in cluster_files:
                        print('\t'+fn)
                cluster_file = cluster_files[0]
            else:
                cluster_file = ''
        if cluster_file is not None:
            f.write('Input File: {}\n'.format(cluster_file))
        # Inference Mode
        mode = maps.mode.upper()
        if mode is None and auto:
            if 'potential' in maps.features:
                mode = 'DV'
            elif 'force' in maps.features:
                mode = 'DF'
            elif 'drift' in maps.features:
                mode = 'DD'
            else:
                mode = 'D'
        if mode is not None:
            f.write('Inference Mode: {}\n'.format(mode))
        # Localization Error
        err = maps.localization_error
        if err is None and auto:
            err = .03
        if err is not None:
            f.write('Localization Error: {:f} [nm]\n'.format(err*1e3))
        # Number of Zones
        f.write('Number of Zones: {}\n'.format(len(zones)))
        # Diffusion Prior
        dp = maps.diffusivity_prior
        if dp is None and auto:
            dp = 0.
        if dp is not None:
            f.write('Diffusion Prior: {:f}\n'.format(dp))
        # Potential Prior
        vp = maps.potential_prior
        if vp is None and auto:
            vp = 0.
        if vp is not None:
            f.write('Potential Prior: {:f}\n'.format(vp))
        # Jeffreys' Prior
        jp = maps.jeffreys_prior
        if jp is None and auto:
            jp = 0
        if jp is not None:
            f.write("Jeffreys' Prior: {:d}\n".format(jp))
        # table header
        f.write('\nZone ID\tTranslocations\tx-Centre\ty-Centre')
        na = np.isnan(maps.maps.values) | np.isinf(maps.maps.values)
        ok = ~np.all(na, axis=1)
        data0 = [ [i, len(zones[i]), zones[i].center] for i in maps.maps.index[ok] ]
        data1 = []
        nvars = 0
        if 'diffusivity' in maps.features:
            f.write('\tDiffusion')
            data1.append(maps['diffusivity'].values[ok])
            nvars += 1
        if 'force' in maps.features:
            f.write('\tx-Force\ty-Force')
            data1.append(maps['force'].values[ok])
            nvars += 2
        if 'potential' in maps.features:
            f.write('\tPotential')
            data1.append(maps['potential'].values[ok])
            nvars += 1
        # table data
        data1 = np.hstack(data1)
        fmt0 = '\n{:d}\t{:d}\t{:f}\t{:f}'
        for r, s in zip(data0, data1):
            i, n, r = r
            s = [ 'NA' if np.isnan(x) or np.isinf(x) else x for x in s ]
            fmt1 = '\t{' + '}\t{'.join('' if x == 'NA' else ':f' for x in s) + '}'
            f.write((fmt0+fmt1).format(i, n, *(r.tolist() + s)))


def export_file(rwa_file, label, cluster_file=None, vmesh_file=None, eps=None, **kwargs):
    if not (cluster_file or vmesh_file):
        raise ValueError('output filename(s) not defined')
    if not label:
        raise ValueError('labels are required')
    vlabel = None
    if isinstance(label, (tuple, list)):
        if label[1:]:
            clabel, vlabel = label
        else:
            clabel = label[0]
    else:
        clabel = label
    if vlabel is None and vmesh_file:
        raise ValueError('label for vmesh file is required')

    analyses = load_rwa(rwa_file)

    if vmesh_file:

        cells, maps = find_artefacts(analyses, ((CellStats, Distributed), Maps), (clabel, vlabel))

    elif cluster_file:

        cells, = find_artefacts(analyses, ((CellStats, Distributed), ), (clabel, ))

    if cluster_file:

        if eps is not None:
            neighbours_kwargs = kwargs.get('neighbours_kwargs', {})
            neighbours_kwargs['eps'] = eps
            kwargs['neighbours_kwargs'] = neighbours_kwargs

        export_to_cluster_file(cells, cluster_file, **kwargs)

    if vmesh_file:

        export_to_vmesh_file(cells, maps, vmesh_file, cluster_file, **kwargs)



def main():
    parser = argparse.ArgumentParser(prog='inferencemap',
        description='InferenceMAP-TRamWAy file converter.')
    parser.add_argument('-t', '--trajectories', metavar='FILE', help='path to trajectory file')
    parser.add_argument('-c', '--cluster', metavar='FILE', help='path to cluster file')
    parser.add_argument('-v', '--vmesh', metavar='FILE', help='path to vmesh file')
    parser.add_argument('-l', '-L', '--label', metavar='LABEL', action='append', default=[], help='labels; the first one identifies the cluster data in the rwa file, the second one identifies the vmesh data') # ideally '-l' is to import, '-L' to export
    import_or_export = parser.add_mutually_exclusive_group()
    import_or_export.add_argument('-o', '--output', metavar='FILE', help='path to the rwa file which to import InferenceMAP files to')
    import_or_export.add_argument('-i', '--input', metavar='FILE', help='path to the rwa file which to export from')
    parser.add_argument('--eps', '--epsilon', metavar='EPSILON', type=float, help='margin for half-space gradient calculation (cluster file exports only; see also `tramway.inference.base.neighbours_per_axis`)')
    parser.add_argument('--auto', action='store_true', help='infer default values for missing parameters')
    args = parser.parse_args()
    kwargs = {}
    if args.input:
        export_file(args.input, args.label, args.cluster, args.vmesh,
            eps=args.eps, auto=args.auto)
    else:
        if args.trajectories:
            root = args.trajectories
        elif args.output and os.path.exists(args.output):
            root = args.output
        else:
            raise ValueError('trajectory file not defined')
        import_file(root, args.label, args.cluster, args.vmesh, args.output)


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from abc import ABCMeta, abstractmethod

class GenericAttribute(metaclass=ABCMeta):
    @property
    @abstractmethod
    def initialized(self):
        """most sub-attributes can be set only once the attribute object is initialized."""
        pass
    @property
    @abstractmethod
    def reified(self):
        """some sub-attributes can be set only as long as the attribute object has not been reified."""
        pass

class Attribute(GenericAttribute):
    @property
    def initialized(self):
        return True




# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module is where features extractible from a single random walk are
defined.
"""

import numpy as np
import pandas as pd
import scipy
import scipy.spatial
import scipy.stats
from scipy.optimize import root_scalar as root_finder

from .visualization import plot_convex_hull
from .rw_misc import rw_is_useless


def _regularize_idminmax(id_min, id_max, N):
    """Helper function ; makes sure that 0 <= id_min < id_max <= N and return
    id_min=0 and id_max=N if those variables are None.
    """
    if id_min is None:
        id_min = 0
    else:
        if id_min < 0:
            raise ValueError(f'index min {id_min} < 0 !')
        elif id_min > N-1:
            raise ValueError(f'index min {id_min} > {N-1} !')
    if id_max is None:
        id_max = N
    else:
        if id_max < 0:
            raise ValueError(f'index min {id_max} < 0 !')
        elif id_max > N:
            raise ValueError(f'index min {id_max} > {N} !')
    if id_max - id_min < 1:
        raise ValueError(f'index min : {id_min}, index max : {id_max} ==> '
                         'empty slice !')
    return id_min, id_max


def _escape_rate(future_dists, mean_step, x):
    """Computes the proportion of times the random walk was able to escape a
    ball centered on the current position with radius x times mean_step.
    For each time step i, we evaluate the maximum distance between future
    positions at time steps j > 0. If this distance is greater than
    x * mean_step, time step i accounts for 1 in the mean. Else it accounts for
    0.

    Parameters
    ----------
    future_dists : N*N ndarray of distances between the random walk at times i
        and j. The lower tridiagonal matrix is null.
    mean_step : scalar, the mean step size of the random walk.
    x : scalar.

    Returns
    -------
    float, between 0 and 1, the proportion of times the random walk escaped.
    """
    return np.count_nonzero(np.argmax(future_dists > x * mean_step,
                                      axis=1)) / future_dists.shape[0]


def _zero_escape_quantile(x, r, future_dists, mean_step):
    """Helper function that is positive if the escape rate (see above function)
    is superior to the scalar r.
    """
    return _escape_rate(future_dists, mean_step, x) - r


def _Dq(Dabs, q, n_R=10, Rmax=0.5):
    """Computes the fractal spectrum function evaluated at q.
    Source : Effective multifractal spectrum of a random walk,
             Berthelsen et al., 1994, equation 1 of related paper.
    """
    Rmax = min(Rmax, np.max(Dabs))
    Rmin = Rmax / 4
    # Log x samples despite linear fit : because we attach more confidence to
    # low R estimates.
    Rs = np.geomspace(Rmin, Rmax, num=n_R)
    Rs = Rs[(Rs != 0) & (Rs != 1)]
    DqR = [(np.log(R) / (q-1) *
            np.log(np.mean(np.mean(Dabs < R, axis=1)**(q-1))))
           for R in Rs]
    fitprms = scipy.stats.linregress(Rs, DqR)
    return fitprms[1]


def _ergodicity_estimator(X):
    """Ergodicity estimation, close to 0 for ergodic processes
    Source : Ergodicity breaking on the neuronal surface emerges from
             random switching between diffusive states, Weron et al., 2017.
    """
    diff = (X[1:] - X[:-1])
    denom = np.sqrt(np.mean(diff**2))
    denom = 1 if denom < 1e-8 else denom
    V = diff / denom
    Cst = np.abs(np.mean(np.exp(1j * V)))**2
    if len(V) > 1:
        n_array = np.unique(np.linspace(
            1, len(V)-1, min(len(V), 7)).astype(int))
        E = np.array([(np.mean(np.exp(1j * (V[n:] - V[:-n]))) -
                       Cst) if n > 0 else 1 - Cst
                      for n in n_array])
        return np.mean(E)
    else:
        return np.nan


def _ergodicity_estimator_v2(X, w=2):
    """Ergodicity estimation, close to 0 for ergodic processes
    Source : Ergodicity breaking on the neuronal surface emerges from
             random switching between diffusive states, Weron et al., 2017.
    """
    N = len(X)
    Cst = 1/N - 1/(N*(N+1)) * np.abs(np.mean(np.exp(1j * w * (X - X[0]))))**2
    n_array = np.unique(np.linspace(1, N-1, min(N, 7)).astype(int))
    E = np.array([np.mean(np.exp(1j * w * (X[n:] - X[:-n]))) + Cst
                  for n in n_array])
    return np.mean(E)


class RandomWalk():
    """Class which helps compute features of random walks.

    Parameters
    ----------
    RW_df : pandas DataFrame of the time - position of the random walk
    zero_time : bool, optional. Whether to make sure that the starting point
        time is 0 (features expect the random walk to start at 0).

    Attributes
    ----------
    data : copy of RW_df
    dims : set, dimensions of the random walk (in {'x', 'y', 'z'})
    length : size of the random walk
    position : numpy values of the positions of the random walk
    t : numpy values of the times of the random walk
    dt_vec : numpy vector of time steps
    is_dt_cst : bool, if time steps are cst
    dt : float, first time step
    Dvec : numpy array of size length * length * len(dims).
        Element (i,j) is the vector position[i] - position[j]
    Dabs : numpy array of size length * length
        Element (i,j) is the distance between position[i] and position[j]
    """

    def __init__(self, RW_df, zero_time=False, check_useless=True,
                 nb_pos_min=3, jump_max=10):
        self.rw_is_useless = (rw_is_useless(RW_df, nb_pos_min, jump_max) if
                              check_useless else False)
        if not self.rw_is_useless or not check_useless:
            self.data = RW_df
            self.dims = set(RW_df.columns).intersection({'x', 'y', 'z'})
            self.length = len(self.data)
            self.position = self.data.loc[:, list(self.dims)].values
            self.t = self.data.t.values
            if zero_time:
                self.t -= self.data.t.min()
            self.dt_vec = self.t[1:] - self.t[:-1]
            self.is_dt_cst = (np.var(self.dt_vec) < 1e-10)
            self.dt = self.dt_vec[0]
            self.Dvec = (self.position[:, np.newaxis] -
                         self.position[np.newaxis, :])
            self.Dabs = np.linalg.norm(self.Dvec, axis=2)

    def __len__(self):
        return self.length

    # Those functions are used to extract sub attributes of the random walks.
    # Will be used when using time windowing feature extraction.
    def get_sub_time(self, id_min, id_max):
        id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
        return self.t[id_min:id_max]

    def get_sub_position(self, id_min, id_max):
        id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
        return self.position[id_min:id_max]

    def get_sub_Dabs(self, id_min, id_max):
        id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
        return self.Dabs[id_min:id_max, id_min:id_max]

    def get_sub_Dvec(self, id_min, id_max):
        id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
        return self.Dvec[id_min:id_max, id_min:id_max]

    def get_sub_steps(self, id_min, id_max):
        id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
        return np.diagonal(self.Dabs, offset=1)[id_min: id_max]

    # Features

    def feat_step(self, id_min=None, id_max=None):
        """Computes statistics on the step size.

        Parameters
        ----------
        RW : RandomWalk object, with size N
        imin : int, should be between 0 and N-2
        imax : int, should be between 2 and N

        Returns
        -------
        dict of moments of the random variable of the size of single steps.
        """
        steps = self.get_sub_steps(id_min, id_max)
        moments = scipy.stats.describe(steps, ddof=0)
        return {'step_mean': moments.mean, 'step_var': moments.variance,
                'step_skewness': moments.skewness,
                'step_kurtosis': moments.kurtosis,
                'step_min': moments.minmax[0], 'step_max': moments.minmax[1]}

    def temporal_msd(self, id_min=None, id_max=None,
                     n_samples=30, sampling='log', use_all=False):
        """Computes the temporal mean squared deviation of the random walk
        between indices id_min and id_max.

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N
        n_samples : optional, number of samples where we compute the temporal
            averaged mean squared displacement.
        sampling : string, optional ; if 'log', log spacing of taus. Shown to
            increase the accuracy of the alpha / diffusion parameters.
        use_all : bool, optional ;
            if True, max tau is N
            else, max_tau is min(N, max(N/2, 10))
            use_all = False increases the accuracy as high n values are less
            precise.

        Returns
        -------
        tau : times at which we computed the mean squared displacement
        msd : mean squared displacement
        """
        subDabs = self.get_sub_Dabs(id_min, id_max)
        # To avoid (if possible) using too many taus with little nb of points
        if use_all:
            max_n = len(subDabs)
        else:
            max_n = min(len(subDabs), max(len(subDabs)/2, 10))
        if sampling == 'log':
            tau_int = np.unique(np.geomspace(1, max_n, num=n_samples,
                                             endpoint=False).astype(int))
        else:
            tau_int = np.unique(np.linspace(1, max_n, num=n_samples,
                                            endpoint=False).astype(int))
        msd = np.array([np.mean(np.diagonal(subDabs, offset=i)**2)
                        for i in tau_int])
        return tau_int, msd

    def feat_msd(self, id_min=None, id_max=None, n_samples=30,
                 sampling='log', use_all=False):
        """Computes the alpha and diffusion (D) coefficients defined by the
        relation between the temporal averaged mean squared displacement and
        time :
        msd(dt) = 4 * D * dt ** (alpha)
        The temporal averaged mean squared displacement is a function defined
        as :
            msd(\tau) = 1/(T-\tau) int_{0}^{T-\tau} (X(t+\tau) - X(t))^2
        For a Brownian motion, msd(dt) = 4 * D * dt

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N
        n_samples : optional, number of samples where we compute the temporal
            averaged mean squared displacement.
        """
        tau, msd = self.temporal_msd(id_min, id_max, n_samples,
                                     sampling, use_all)
        tau = tau * self.dt
        if len(tau) > 1:
            slope, intercept, r, _, _ = scipy.stats.linregress(np.log10(tau),
                                                               np.log10(msd))
            return {'msd_alpha': slope, 'msd_rval': r,
                    'msd_diffusion': 10**intercept / 4}
        else:
            return {'msd_alpha': np.nan, 'msd_rval': np.nan,
                    'msd_diffusion': np.nan}

    def feat_drift(self, id_min=None, id_max=None, raw_val=False, **kwargs):
        """Computes the mean drift norm of the random walk, defined as ||alpha||
        where X(t) = alpha * t.

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N

        Returns
        -------
        dict of the drift norm value.
        """
        subposition = self.get_sub_position(id_min, id_max)
        subt = self.get_sub_time(id_min, id_max)
        drift = np.mean(((subposition[1:] - subposition[0]) /
                         subt[1:, np.newaxis]), axis=0)
        val = np.linalg.norm(drift)
        if raw_val:
            return val
        else:
            return {'drift_norm': val}

    def stats_pdf(self, id_min=None, id_max=None, n_samples=30):
        """Computes moments of the distribution of distances
            ||X(t+tau) - X(t)||, for n_samples different taus.

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N
        n_samples : optional, number of samples where we compute the moments of
            the distribution of distances.

        Returns
        -------
        tau : taus for which we computed the moments of ||X(t+tau) - X(t)||.
        pdf_stats : array of shape n_samples * 4.
        """
        subDabs = self.get_sub_Dabs(id_min, id_max)
        tau_int = np.unique(np.geomspace(1, len(subDabs)-1, num=n_samples,
                                         endpoint=False).astype(int))
        pdf_stats = [scipy.stats.describe(np.diagonal(subDabs, offset=i),
                                          ddof=0) for i in tau_int]
        pdf_stats = np.array([[x.mean, x.variance, x.skewness, x.kurtosis]
                              for x in pdf_stats])
        # pdf_stats = np.stack(([np.mean(np.diagonal(subDabs, offset=i))
        #                        for i in tau_int],
        #                       [np.var(np.diagonal(subDabs, offset=i))
        #                        for i in tau_int])).T
        return tau_int * self.dt, pdf_stats

    def feat_pdf(self, id_min=None, id_max=None, n_samples=30):
        """Returns features extracted from the evolution of the moments of
        ||X(t+tau) - X(t)|| with tau.
        Those features are :
            - for the mean and variance :
                - we fit alpha, beta such as mean(tau) = beta * tau ** beta
                - we return alpha, beta and the correlation r.
            - for the skewness and kurtosis :
                - we return the mean and variance accross taus.

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N
        n_samples : optional, number of samples where we compute the moments of
            the distribution of distances.

        Returns
        -------
        dictionary of features extracted from the estimated PDF of the random
            walk.
        """
        tau, pdf_stats = self.stats_pdf(id_min, id_max, n_samples)
        keys = ['pdf_alpha_mean', 'pdf_beta_mean', 'pdf_rval_mean',
                'pdf_alpha_var', 'pdf_beta_var', 'pdf_rval_var',
                'pdf_mean_skewness', 'pdf_var_skewness',
                'pdf_mean_kurtosis', 'pdf_var_kurtosis']
        # keys = ['pdf_alpha_mean', 'pdf_beta_mean', 'pdf_rval_mean',
        #         'pdf_alpha_var', 'pdf_beta_var', 'pdf_rval_var']
        if len(tau) > 1:
            fit_mean = scipy.stats.linregress(np.log10(tau),
                                              np.log10(pdf_stats[:, 0]))
        else:
            fit_mean = np.array([np.nan, np.nan, np.nan])
        if len(pdf_stats[:, 1][pdf_stats[:, 1] > 0]) > 1:
            x = np.log10(tau[pdf_stats[:, 1] > 0])
            y = np.log10(pdf_stats[:, 1][pdf_stats[:, 1] > 0])
            fit_var = scipy.stats.linregress(x, y)
        else:
            fit_var = np.array([np.nan, np.nan, np.nan])
        mu_sk, var_sk = np.mean(pdf_stats[:, 2]), np.var(pdf_stats[:, 2])
        mu_ku, var_ku = np.mean(pdf_stats[:, 3]), np.var(pdf_stats[:, 3])
        vals = [fit_mean[0], 10**fit_mean[1], fit_mean[2],
                fit_var[0], 10**fit_var[1], fit_var[2],
                mu_sk, var_sk, mu_ku, var_ku]
        # vals = [fit_mean[0], 10**fit_mean[1], fit_mean[2],
        #         fit_var[0], 10**fit_var[1], fit_var[2]]
        return dict(list(zip(keys, vals)))

    def convex_hull(self, id_min=None, id_max=None, display=False,
                    only_area=False, **kwargs):
        """Returns the area, perimeter and maximum distance between 2 positions
        of the random walk.
        Makes used of the scipy.spatial function ConvexHull.

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N
        """
        try:
            subposition = self.get_sub_position(id_min, id_max)
            subposition = np.unique(subposition, axis=0)
            hull = scipy.spatial.ConvexHull(subposition)
            area, perimeter = hull.volume, hull.area
            if only_area:
                return area
            else:
                hull_points = subposition[hull.vertices]
                n = len(hull_points)
                DX = np.broadcast_to(hull_points[:, 0], (n, n))
                DY = np.broadcast_to(hull_points[:, 1], (n, n))
                D = (DX - DX.T)**2 + (DY - DY.T)**2
                imax = np.argmax(D)
                max_dist = np.sqrt(np.max(D))
                if display:
                    plot_convex_hull(subposition, hull, imax)
                return area, perimeter, max_dist
        except:
            return np.nan, np.nan, np.nan

    def gyration_tensor(self, id_min=None, id_max=None):
        X = self.get_sub_position(id_min, id_max)
        Xm = np.mean(X, axis=0)
        a, b = np.mean((X - Xm[np.newaxis, :])**2, axis=0)
        c = np.mean((X[:, 0] - Xm[0]) * (X[:, 1] - Xm[1]))
        return np.array([[a, c], [c, b]])

    def asphericity(self, id_min=None, id_max=None, **kwargs):
        try:
            X = self.get_sub_position(id_min, id_max)
            v = X - np.mean(X, axis=0)
            T = v.T @ v / v.shape[1]
            eig_vals = np.linalg.eigvals(T)
            num = (eig_vals[1] - eig_vals[0])**2
            denom = np.sum(eig_vals)**2
            return np.real(num / denom)
        except:
            return np.nan

    def asymmetry(self, id_min=None, id_max=None):
        """Returns a feature which controls how assymetric the random walk is.
        May help to detect drift.
        """
        try:
            T = self.gyration_tensor(id_min, id_max)
            l1, l2 = np.linalg.eigvals(T)
            return - np.log(1 - (l1 - l2)**2 / (2 * (l1 + l2)**2))
        except:
            return np.nan

    def efficiency(self, id_min=None, id_max=None):
        """Returns a feature which controls how efficient the random walk was
        at going from the starting point to the ending point.
        May help to detect drift.
        """
        try:
            subDabs = self.get_sub_Dabs(id_min, id_max)
            num = subDabs[0, -1]**2
            denom = np.sum(np.diagonal(subDabs, offset=1)**2)
            return num / denom
        except:
            return np.nan

    def kurtosis(self, id_min=None, id_max=None, **kwargs):
        """Source : J. A. Helmuth, C. J. Burckhardt, P. Koumoutsakos,
        U. F. Greber, and I. F. Sbalzarini, Journal of Structural Biology 159,
        347 (2007).
        """
        try:
            X = self.get_sub_position(id_min, id_max)
            T = self.gyration_tensor(id_min, id_max)
            w, v = np.linalg.eig(T)
            r = v[:, np.argmax(w)]
            xp = np.sum(X * r, axis=1)
            xpm, xpstd = np.mean(xp), np.std(xp)
            return np.mean((xp - xpm)**4) / xpstd**4
        except:
            return np.nan

    def straightness(self, id_min=None, id_max=None):
        """Returns straightness, much like efficiency feature, but with
        absolute distances (not squared).
        """
        try:
            subDabs = self.get_sub_Dabs(id_min, id_max)
            rs = np.diagonal(subDabs, offset=1)
            return subDabs[0, -1] / np.sum(rs)
        except:
            return np.nan

    def feat_shape(self, id_min=None, id_max=None):
        """Regroups all features related to the shape of the random walk into
        a single function.

        Parameters
        ----------
        id_min : optional, int should be between 0 and N-1
        id_max : optional, int should be between 1 and N
        """
        id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
        keys = ['area', 'perimeter', 'max_dist', 'asymmetry', 'asphericity',
                'efficiency', 'kurtosis', 'straightness']
        area, perimeter, max_dist = self.convex_hull(id_min, id_max)
        asphericity = self.asphericity(id_min, id_max)
        asymmetry = self.asymmetry(id_min, id_max)
        efficiency = self.efficiency(id_min, id_max)
        kurtosis = self.kurtosis(id_min, id_max)
        straightness = self.straightness(id_min, id_max)
        vals = [area, perimeter, max_dist, asymmetry, asphericity,
                efficiency, kurtosis, straightness]
        return dict(list(zip(keys, vals)))

    def feat_escape_time(self, id_min=None, id_max=None, method='brenth'):
        """Returns features derived from the escape time : the time at which
        we escape a distance x.
        Those features are q_i, i in {25, 50, 75}, where q_i is the distance
        from which i% of starting positions escaped from.
        """
        subDabs = self.get_sub_Dabs(id_min, id_max)
        mean_step = np.mean(np.diagonal(subDabs, offset=1))
        sample = np.random.permutation(len(subDabs))[:min(100, len(subDabs))]
        sample = np.sort(sample)
        subDabs = subDabs[sample][:, sample]
        subDfuture = np.triu(subDabs)
        keys = ['escape_dist_q1', 'escape_dist_median', 'escape_dist_q3']
        qs = [0.25, 0.5, 0.75]
        d_qs = []
        for q in qs:
            # We have to check if the random walk is not too much immobile
            # to make sure that at least q% of points finally move, so that
            # the solver has 2 a=0 and b such that f(a) > 0 (and f(b) > 0)
            frac_move = (np.unique(subDfuture, axis=1).shape[1] /
                         subDfuture.shape[0])
            if q < frac_move:
                try:
                    args = (q, subDfuture, mean_step)
                    sol = root_finder(_zero_escape_quantile, args=args,
                                      method=method, bracket=(0, 100))
                    d_qs.append(sol.root)
                except:
                    d_qs.append(0)
            else:
                d_qs.append(0)
        vals = np.array(d_qs) * mean_step
        return dict(list(zip(keys, vals)))

    def feat_angle_old(self, id_min=None, id_max=None):
        """Returns moments related to the distribution of angles.
        """
        subvec = np.diagonal(self.get_sub_Dvec(id_min, id_max),
                             offset=-1)
        vecnorm = np.linalg.norm(subvec, axis=0)
        no_mvt = vecnorm < 1e-6
        vecnorm[no_mvt] = 1
        unit_vec = subvec / vecnorm[np.newaxis, :]
        keys = ['angle_mean', 'angle_var', 'angle_skewness',
                'angle_kurtosis', 'angle_min', 'angle_max']
        if unit_vec.shape[1] > 1:
            prod = np.sum(unit_vec[:, :-1] * unit_vec[:, 1:], axis=0)
            angles = np.arccos(np.clip(prod, -1.0, 1.0))
            undefined_angle = no_mvt[:-1]
            if len(undefined_angle) > 0:
                undefined_angle[-1] *= no_mvt[-1]
            nb_undefined = np.sum(undefined_angle)
            angles[undefined_angle] = np.random.uniform(0, np.pi,
                                                        size=nb_undefined)
            moments = scipy.stats.describe(angles, ddof=0)
            vals = [moments.mean, moments.variance, moments.skewness,
                    moments.kurtosis, moments.minmax[0], moments.minmax[1]]
            return dict(list(zip(keys, vals)))
        else:
            return dict(list(zip(keys, np.nan * np.ones(len(keys)))))

    def feat_angle(self, id_min=None, id_max=None):
        """Returns variance, and first two autocorrelations of angles.
        """
        subvec = np.diagonal(self.get_sub_Dvec(id_min, id_max),
                             offset=-1)
        vecnorm = np.linalg.norm(subvec, axis=0)
        no_mvt = vecnorm < 1e-6
        vecnorm[no_mvt] = 1
        unit_vec = subvec / vecnorm[np.newaxis, :]
        keys = ['angle_var', 'angle_autocorr1', 'angle_autocorr2']
        if unit_vec.shape[1] > 1:
            prod = np.sum(unit_vec[:, :-1] * unit_vec[:, 1:], axis=0)
            angles = np.arccos(np.clip(prod, -1.0, 1.0))
            undefined_angle = no_mvt[:-1]
            if len(undefined_angle) > 0:
                undefined_angle[-1] *= no_mvt[-1]
            nb_undefined = np.sum(undefined_angle)
            angles[undefined_angle] = np.random.uniform(0, np.pi,
                                                        size=nb_undefined)

            mean_angle, var_angle = np.mean(angles), np.var(angles)
            autocorrs_angle = [np.nan, np.nan]
            if var_angle > 0:
                for i in range(1, 3):
                    if len(angles) > i:
                        autocov_angle = np.mean((angles[i:] - mean_angle) *
                                                (angles[:-i] - mean_angle))
                        autocorrs_angle[i-1] = autocov_angle / var_angle
            vals = [var_angle] + autocorrs_angle
            return dict(list(zip(keys, vals)))
        else:
            return dict(list(zip(keys, np.nan * np.ones(len(keys)))))

    def autocorr(self, tau_int, id_min=None, id_max=None, n_samples=30):
        steps = self.get_sub_steps(id_min, id_max)
        n = len(steps)
        mean_step, var_step = np.mean(steps), np.var(steps)
        if len(steps) > tau_int:
            return np.mean((steps[tau_int:] - mean_step) *
                           (steps[:-tau_int] - mean_step)) / var_step
        else:
            return np.nan

    def feat_step_autocorr(self, id_min=None, id_max=None):
        """Returns moments of the distribution of the autocorrelation function
        defined as :
            corr(tau) = E_t[(d(t) - E[d])(d(t+tau) - E[d])] / Var(d(t))
        with d the vector of single step distances.
        """
        steps = self.get_sub_steps(id_min, id_max)
        n = len(steps)
        mean_step, var_step = np.mean(steps), np.var(steps)
        keys = ['autocorr_1', 'autocorr_2', 'autocorr_3']
        autocorrs = np.ones(3) * np.nan
        if var_step > 0:
            for i in range(1, 4):
                if len(steps) > i:
                    autocov = np.mean((steps[i:] - mean_step) *
                                      (steps[:-i] - mean_step))
                    autocorrs[i-1] = autocov / var_step
        return dict(list(zip(keys, autocorrs)))

    def feat_fractal_spectrum(self, id_min=None, id_max=None,
                              n_samples=30, nR=20, qs=None):
        """Computes a feature derived from the fractal spectrum function
        evaluated at q. This feature is the derivative of the function at q=0.
        Source : Effective multifractal spectrum of a random walk,
                Berthelsen et al., 1994, equation 1 of related paper.
        """
        subDabs = self.get_sub_Dabs(id_min, id_max)
        n = len(subDabs)
        chosen_points = np.sort(np.random.permutation(n)[:min(n, n_samples)])
        subDabs_sampled = subDabs[chosen_points]
        if qs is None:
            try:
                D_inf = _Dq(subDabs_sampled, -1.5, n_R=nR)
                D_sup = _Dq(subDabs_sampled, 1.5, n_R=nR)
                Dqgrad0 = 1/3 * (D_sup - D_inf)
                return {'frac_grad0': Dqgrad0}
            except:
                return {'frac_grad0': np.nan}
        else:
            Dq = [_Dq(subDabs_sampled, q, n_R=nR) for q in qs]
            return Dq

    def feat_ergodicity(self, id_min=None, id_max=None):
        """Ergodicity estimation features, close to 0 for ergodic processes
        Source : Ergodicity breaking on the neuronal surface emerges from
                random switching between diffusive states, Weron et al., 2017.
        """
        subposition = self.get_sub_position(id_min, id_max)
        vecs = [subposition[:, i] for i in range(len(self.dims))]
        Fs1 = np.abs(np.array(list(map(_ergodicity_estimator, vecs))))
        Fs2 = np.abs(np.array(list(map(_ergodicity_estimator_v2, vecs))))
        type1 = {f'ergo_{dim}': Fs1[i] for i, dim in enumerate(self.dims)}
        type2 = {f'ergo2_{dim}': Fs2[i] for i, dim in enumerate(self.dims)}
        return {**type1, **type2}

    def feat_gaussianity(self, id_min=None, id_max=None, n_samples=30):
        """Gaussianity estimation.
        Source : Ernst, D., Köhler, J., & Weiss, M. (2014). Probing the type of
        anomalous diffusion with single-particle tracking.
        Physical Chemistry Chemical Physics, 16(17), 7686-7691.
        """
        subDabs = self.get_sub_Dabs(id_min, id_max)
        ns = np.unique(np.linspace(1, len(subDabs)-1, n_samples).astype(int))
        gauss_n = np.zeros(len(ns))
        for i, n in enumerate(ns):
            d_n = np.diagonal(subDabs, offset=n)
            gauss_n[i] = np.mean(d_n**4) / ((np.mean(d_n**2)**2)) - 1
        grad_gauss_n = np.gradient(gauss_n) if len(gauss_n) > 1 else np.nan
        mean_grad = np.mean(grad_gauss_n)
        mean_gauss = np.mean(gauss_n)
        return {'gaussian_grad': mean_grad, 'gaussian_mean': mean_gauss}

    def get_all_features(self, id_min=None, id_max=None,
                         n_samples=30, old=False, time_evol=False, **kwargs):
        if self.rw_is_useless:
            return {}
        else:
            if old:
                func_feat_ang = self.feat_angle_old
            else:
                func_feat_ang = self.feat_angle
            id_min, id_max = _regularize_idminmax(id_min, id_max, self.length)
            meta_info = {
                'size': id_max - id_min,
                'dt': self.dt,
                'is_dt_cst': self.is_dt_cst,
                't_min': self.t[id_min],
                't_max': self.t[id_max-1]
            }
            feats = {**meta_info,
                     **func_feat_ang(id_min, id_max),
                     **self.feat_drift(id_min, id_max),
                     **self.feat_ergodicity(id_min, id_max),
                     **self.feat_escape_time(id_min, id_max),
                     **self.feat_fractal_spectrum(id_min, id_max,
                                                  n_samples=n_samples, nR=4),
                     **self.feat_gaussianity(id_min, id_max,
                                             n_samples=n_samples),
                     **self.feat_msd(id_min, id_max, n_samples=n_samples),
                     **self.feat_pdf(id_min, id_max, n_samples=n_samples),
                     **self.feat_shape(id_min, id_max),
                     **self.feat_step(id_min, id_max),
                     **self.feat_step_autocorr(id_min, id_max)}
            if time_evol:
                feats.update(get_features_time(self, **kwargs))
            return feats

    def get_features_vector(self, func, w, step, **kwargs):
        features_time = {}
        w2 = w / 2
        w2i = int(w2)
        id_f = (self.length - w2i - 1) - (1 if w % 2 == 1 else 0)
        for i in range(0, self.length - w, step):
            features_time[(i + w2i) * self.dt] = func(self, i, i + w, **kwargs)
        for i in range(w2i):
            features_time[i * self.dt] = features_time[w2i * self.dt]
            id_i = (self.length - i - 1)
            features_time[id_i * self.dt] = features_time[id_f * self.dt]
        return pd.DataFrame.from_dict(features_time, orient='index')


def feat_step(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_step(id_min=id_min, id_max=id_max)


def feat_msd(RW, id_min=None, id_max=None, n_samples=30, **kwargs):
    return RW.feat_msd(id_min=id_min, id_max=id_max, n_samples=n_samples)


def temporal_msd(RW, id_min=None, id_max=None,
                 n_samples=30, sampling='log', use_all=False):
    return RW.temporal_msd(id_min=id_min, id_max=id_max,
                           n_samples=n_samples, sampling='log', use_all=False)


def feat_drift(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_drift(id_min=id_min, id_max=id_max, **kwargs)


def feat_pdf(RW, id_min=None, id_max=None, n_samples=30, **kwargs):
    return RW.feat_pdf(id_min=id_min, id_max=id_max, n_samples=n_samples)


def convex_hull(RW, id_min=None, id_max=None, display=False, **kwargs):
    return RW.convex_hull(id_min=id_min, id_max=id_max, display=display,
                          **kwargs)


def kurtosis(RW, id_min=None, id_max=None, **kwargs):
    return RW.kurtosis(id_min=id_min, id_max=id_max, **kwargs)


def asphericity(RW, id_min=None, id_max=None, **kwargs):
    return RW.asphericity(id_min=id_min, id_max=None, **kwargs)


def feat_shape(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_shape(id_min=id_min, id_max=id_max)


def feat_escape_time(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_escape_time(id_min=id_min, id_max=id_max)


def feat_angle_old(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_angle(id_min=id_min, id_max=id_max)


def feat_angle(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_angle(id_min=id_min, id_max=id_max)


def feat_step_autocorr(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_step_autocorr(id_min=id_min, id_max=id_max)


def feat_fractal_spectrum(RW, id_min=None, id_max=None, n_samples=30,
                          nR=20, qs=None, **kwargs):
    return RW.feat_fractal_spectrum(id_min=id_min, id_max=id_max,
                                    n_samples=n_samples, nR=nR, qs=qs)


def feat_ergodicity(RW, id_min=None, id_max=None, **kwargs):
    return RW.feat_ergodicity(id_min=id_min, id_max=id_max)


def feat_gaussianity(RW, id_min=None, id_max=None, n_samples=30, **kwargs):
    return RW.feat_gaussianity(id_min=id_min, id_max=id_max,
                               n_samples=n_samples)


def get_all_features(RW, id_min=None, id_max=None, n_samples=30,
                     old=False, time_evol=False, **kwargs):
    return RW.get_all_features(id_min=id_min, id_max=id_max,
                               n_samples=n_samples, old=old,
                               time_evol=time_evol, **kwargs)


def get_features_vector(RW, func, w, step, **kwargs):
    return RW.get_features_vector(func, w, step, **kwargs)


def mean_through_time(func, RW_obj, n, **kwargs):
    N = len(RW_obj)
    delta_ts = np.unique(np.geomspace(kwargs['start'], N-1, n)).astype(int)
    vals_ts = []
    for dt in delta_ts:
        vals_ts.append(np.mean(np.array([func(RW_obj, i, i+dt, **kwargs)
                                         for i in range(N-dt)])))
    vmax = func(RW_obj, 0, N, **kwargs)
    return delta_ts, np.array(vals_ts)/vmax


def get_features_time(RW, n_t_samples=10, **kwargs):
    N = len(RW)
    Tmax = RW.t.max()
    dict_feat_vals = {}
    times = np.unique(np.geomspace(1, N-1, n_t_samples)).astype(int)
    msd = np.array([np.mean(np.diagonal(RW.Dabs, offset=i)**2)
                    for i in times])
    msd /= Tmax
    dict_feat_vals['msd'] = msd
    times = np.unique(np.geomspace(1, N-2, n_t_samples)).astype(int)
    autocorrs = np.array([RW.autocorr(t, id_min=0, id_max=N, n_samples=30)
                          for t in times])
    dict_feat_vals['autocorrs'] = autocorrs
    ds = RW.get_sub_steps(0, N)
    steps_chosen = np.linspace(0, N-2, 10).astype(int)
    cum_step = np.cumsum(ds / np.sum(ds))[steps_chosen]
    dict_feat_vals['cum_step'] = cum_step
    _, cvhs = mean_through_time(convex_hull, RW, n=n_t_samples, **kwargs)
    dict_feat_vals['cvhs'] = cvhs
    _, kurts = mean_through_time(kurtosis, RW, n=n_t_samples, **kwargs)
    dict_feat_vals['kurts'] = kurts
    _, asphericities = mean_through_time(asphericity, RW,
                                         n=n_t_samples, **kwargs)
    dict_feat_vals['asphericities'] = asphericities
    _, drift = mean_through_time(feat_drift, RW,
                                 n=n_t_samples, **kwargs)
    dict_feat_vals['drift'] = drift
    prms = {}
    for feature, vals in dict_feat_vals.items():
        if feature != 'autocorrs':
            vals = np.log(vals)
            feature += '_log'
        prms.update(dict(list(zip([f'{feature}_{i}'
                                   for i in range(1, n_t_samples+1)],
                                  vals))))
    return prms



# -*- coding: utf-8 -*-

# Copyright © 2018, Alexander Serov

import time


class stopwatch:
    """A class for measuring execution time."""

    def __init__(self, name, verbose=True):
        self.name = name
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = time.time()
        delta = self.end - self.start
        if self.verbose:
            print('\n{name} completed in {t} s.\n'.format(name=self.name, t=round(delta, 1)))


# def stopwatch_dec(func):
#     """An alternative decorator for measuring the elapsed time."""
#
#     def wrapper(*args, **kwargs):
#         start = time.time()
#         results = func(*args, **kwargs)
#         delta = time.time() - start
#         print(f'\n{self.name} completed in {round(delta, 1)} s.\n')
#         return results
#     return wrapper



# TODO: clean-up the file
import sys
from collections import deque  # a standard FIFO queue

import numpy as np


def group_by_sign(cells, tqdm, B_threshold=10, random_group_order=True, **kwargs):
    """
    Parameters:
    lg_Bs,   nd-array of Bayes factor values of individual cells, Nx1
    adjacency,  cell adjacency matrix, sparse, NxN

    Return:

    """

    # Prepare the data. A Cell is a CellStats object
    adjacency = cells.adjacency
    N = cells.adjacency.shape[0]
    lg_Bs = np.full(N, np.nan)
    for key in cells:
        lg_Bs[key] = cells[key].lg_B
    lg_Bs = np.concatenate(lg_Bs, axis=None)

    N = len(lg_Bs)
    # Initialize queues
    # remaining_queue = deque(range(N))
    queue = deque()
    finished = set()
    groups = np.full(N, np.nan)

    signs = np.sign(lg_Bs)

    def process_one_cell(i, group):
        """If cell's neighbors have same sign, add to active queue"""

        # Mark cell
        finished.add(i)
        groups[i] = group

        neighbs = adjacency.getrow(i).nonzero()[1]
        sign = signs[i]
        for n in neighbs:
            if signs[n] == sign:
                queue.append(n)

    # Main cycle
    group = 0
    for i in tqdm(range(N), desc='Grouping force regions'):
        if i in finished:
            continue

        # No group if nan
        if np.isnan(signs[i]):
            continue

        queue.append(i)
        while queue:
            j = queue.popleft()
            if j not in finished:
                process_one_cell(j, group)
        group += 1
    max_group = group - 1

    # Groups identified. Now calculate the Bayes factors of the groups, and the thresholded maps
    group_lg_B = np.full(N, np.nan)
    for group in range(max_group):
        inds = groups == group
        group_lg_B[inds] = lg_Bs[inds].sum()

    # Threshold the results at the given significance level
    group_forces = 1 * (group_lg_B >= np.log10(B_threshold)) - \
        1 * (group_lg_B <= -np.log10(B_threshold))

    # Randomize group order
    if random_group_order:
        new_groups = np.random.permutation(max_group + 1)
        for i in range(len(groups)):
            if ~np.isnan(groups[i]):
                groups[i] = new_groups[int(groups[i])]

    # Store data in cells
    for key in cells:
        cells[key].groups = groups[key]
        cells[key].group_lg_B = group_lg_B[key]
        cells[key].group_forces = group_forces[key]

    # print(groups, group_lg_B, group_forces)
    # sys.exit(1)
    # raise RuntimeError('Manual stop')




import rwa
from . import *
import importlib
import copy

# 0.2 -> 0.3: typo
translation_table = [
        ('tramway.tesselation.base.CellStats',  'tramway.tessellation.base.CellStats'),
        ('tramway.tesselation.base.Delaunay',   'tramway.tessellation.base.Delaunay'),
        ('tramway.tesselation.base.Voronoi',    'tramway.tessellation.base.Voronoi'),
        ('tramway.tesselation.base.RegularMesh',        'tramway.tessellation.base.RegularMesh'),
        ('tramway.tesselation.nesting.NestedTesselations',      'tramway.tessellation.nesting.NestedTessellations'),
        ('tramway.tesselation.kdtree.KDTreeMesh',       'tramway.tessellation.kdtree.KDTreeMesh'),
        ('tramway.tesselation.kmeans.KMeansMesh',       'tramway.tessellation.kmeans.KMeansMesh'),
        ('tramway.tesselation.gas.GasMesh',     'tramway.tessellation.gas.GasMesh'),
        ('tramway.tesselation.time.TimeLattice',        'tramway.tessellation.time.TimeLattice'),
        ('tramway.tesselation.plugins.window.SlidingWindow',    'tramway.tessellation.plugins.window.SlidingWindow'),
        ]

# 0.2 -> 0.3: new project structure
translation_table += [
        ('tramway.inference.diffusivity.DV',    'tramway.inference.dv.DV'),
        ('tramway.spatial.scaler.Scaler',       'tramway.core.scaler.Scaler'),
        ('tramway.tessellation.base.RegularMesh',       'tramway.tessellation.grid.RegularMesh'),
        ('tramway.spatial.dichotomy.Dichotomy', 'tramway.tessellation.kdtree.dichotomy.Dichotomy'),
        ('tramway.spatial.dichotomy.ConnectedDichotomy',        'tramway.tessellation.kdtree.dichotomy.ConnectedDichotomy'),
        ('tramway.spatial.gas.Gas',     'tramway.tessellation.gwr.gas.Gas'),
        ('tramway.spatial.graph.array.ArrayGraph',      'tramway.tessellation.gwr.graph.array.ArrayGraph'),
        ('tramway.tessellation.gas.GasMesh',    'tramway.tessellation.gwr.GasMesh'),
        ('tramway.tessellation.plugins.window.SlidingWindow',   'tramway.tessellation.window.SlidingWindow'),
        ]

# 0.2 -> 0.3: type refinements
translation_table += [
        ('tramway.inference.base.Cell', 'tramway.inference.base.Translocations'),
        ]


def translate_types(translation_table):
        """Translate types for rwa files.
        """
        # reduce the table
        _table = dict(translation_table)
        _any_update = True
        while _any_update:
                _any_update = False
                for _former_type, _current_type in list(_table.items()):
                        try:
                                _more_current_type = _table[_current_type]
                        except KeyError:
                                pass
                        else:
                                _table[_former_type] = _more_current_type
                                _any_update = True

        # load and register
        for _former_type, _current_type in _table.items():
                _module_name = _current_type.split('.')
                _type_name = _module_name.pop()
                _module_name = '.'.join(_module_name)
                _module = importlib.import_module(_module_name)
                _type = getattr(_module, _type_name)
                try:
                        _storable = rwa.hdf5.hdf5_service.byStorableType(_current_type)
                except:
                        _storable = rwa.default_storable(_type, storable_type=_current_type)
                        _storable_handler = _storable.handlers[0]
                else:
                        _storable_handler = copy.copy(_storable.asVersion())
                _storable_handler._poke = None # peek only
                _storable = rwa.Storable(_type, key=_former_type, handlers=_storable_handler)
                rwa.hdf5_storable(_storable, agnostic=True)

translate_types(translation_table)





from . import *
from ..attribute import initializer_method

from_tiff_file  = initializer_method( ImagesInitializer.from_tiff_file  )
from_tiff_files = initializer_method( ImagesInitializer.from_tiff_files )





from . import *
from ..attribute import initializer_method

from_ascii_file    = initializer_method( SPTDataInitializer.from_ascii_file    )
from_ascii_files   = initializer_method( SPTDataInitializer.from_ascii_files   )
from_dataframe     = initializer_method( SPTDataInitializer.from_dataframe     )
from_dataframes    = initializer_method( SPTDataInitializer.from_dataframes    )
from_mat_file      = initializer_method( SPTDataInitializer.from_mat_file      )
from_mat_files     = initializer_method( SPTDataInitializer.from_mat_files     )
from_rwa_file      = initializer_method( SPTDataInitializer.from_rwa_file      )
from_rwa_files     = initializer_method( SPTDataInitializer.from_rwa_files     )
from_rw_generator  = initializer_method( SPTDataInitializer.from_rw_generator  )
from_analysis_tree = initializer_method( SPTDataInitializer.from_analysis_tree )
#from_tracker      = initializer_method( SPTDataInitializer.from_tracker       )





from . import *
from ..attribute import initializer_method

from_single_particle = initializer_method( TrackerInitializer.from_single_particle )
from_non_tracking    = initializer_method( TrackerInitializer.from_non_tracking    )

